from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware
from wexample_wex_core.decorator.option import option

from wexample_wex_addon_app.middleware.suite_or_each_package_middleware import (
    SuiteOrEachPackageMiddleware,
)
from wexample_wex_addon_app.workdir.code_base_workdir import CodeBaseWorkdir

if TYPE_CHECKING:
    from wexample_app.response.boolean_response import BooleanResponse
    from wexample_wex_core.context.execution_context import ExecutionContext


@option(
    name="force",
    type=bool,
    default=False,
    is_flag=True,
    description="Force bump even if package has no new content",
)
@option(name="yes", type=bool, default=False, is_flag=True)
@middleware(middleware=SuiteOrEachPackageMiddleware)
@command(
    type=COMMAND_TYPE_ADDON,
    description="Bump version only if package has new content (HEAD not tagged). Use --force to bump regardless of changes. Use --all-packages to bump all packages, --packages-only to exclude suite, --suite-only to exclude packages.",
)
def app__package__bump(
    context: ExecutionContext,
    app_workdir: CodeBaseWorkdir,
    yes: bool = False,
    force: bool = False,
) -> BooleanResponse:
    from wexample_app.response.boolean_response import BooleanResponse

    package_name = app_workdir.get_package_name()
    bumped = app_workdir.bump(interactive=not yes, force=force)

    if bumped:
        context.io.success(f"Successfully bumped {package_name}.")
    else:
        context.io.log(f"Bump aborted for {package_name}.")

    return BooleanResponse(kernel=context.kernel, content=bumped)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from wexample_app.const.env import ENV_NAME_PROD
from wexample_app.const.globals import APP_PATH_APP_MANAGER
from wexample_config.config_value.config_value import ConfigValue
from wexample_file.helper.line import line_count_recursive
from wexample_filestate.const.types_state_items import TargetFileOrDirectoryType
from wexample_helpers.classes.abstract_method import abstract_method
from wexample_helpers.decorator.base_class import base_class
from wexample_helpers_git.helpers.git import git_has_changes_since_tag
from wexample_prompt.enums.terminal_color import TerminalColor
from wexample_wex_core.workdir.workdir import Workdir

from wexample_wex_addon_app.const.path import APP_PATH_TEST
from wexample_wex_addon_app.workdir.mixin.app_workdir_mixin import AppWorkdirMixin

if TYPE_CHECKING:
    from wexample_config.const.types import DictConfig
    from wexample_filestate.result.file_state_result import FileStateResult

    from wexample_wex_addon_app.workdir.framework_packages_suite_workdir import (
        FrameworkPackageSuiteWorkdir,
    )


@base_class
class BasicAppWorkdir(AppWorkdirMixin, Workdir):
    def app_install(self, env: str | None = None, force: bool = False) -> bool:
        return True

    def apply(
        self,
        force: bool = False,
        scopes=None,
        filter_path: str | None = None,
        filter_operation: str | None = None,
        max: int = None,
        **kwargs,
    ) -> FileStateResult:
        from wexample_filestate.result.file_state_result import FileStateResult
        from wexample_helpers.helpers.repo import repo_get_state, repo_has_changed_since

        # Hash protection is only active when all filter parameters are None
        # to avoid false positives when apply behavior is modified by parameters
        hash_protection_active = (
            scopes is None
            and filter_path is None
            and filter_operation is None
            and max is None
        )

        registry_file = self.get_registry_file()
        registry = registry_file.read_config()
        last_update_hash = registry.search(
            "file_state.last_update_hash"
        ).get_str_or_none()

        if (
            force
            or not hash_protection_active
            or (
                last_update_hash is None
                or repo_has_changed_since(
                    previous_state=last_update_hash, cwd=self.get_path()
                )
            )
        ):
            # Reset hash
            registry.set_by_path("file_state.last_update_hash", None)
            registry_file.write_config()

            result = super().apply(
                scopes=scopes,
                filter_path=filter_path,
                filter_operation=filter_operation,
                max=max,
                **kwargs,
            )

            # Save hash only if protection is active
            if hash_protection_active:
                registry.set_by_path(
                    "file_state.last_update_hash", repo_get_state(cwd=self.get_path())
                )
                registry_file.write_config()

            return result

        self.log("No change since last pass, skipping.", indentation=1)
        return FileStateResult(state_manager=self)

    def bump(self, interactive: bool = False, force: bool = False, **kwargs) -> bool:
        """Create a version-x.y.z branch, update the version number in config. Don't commit changes."""
        from wexample_helpers.helpers.version import version_increment
        from wexample_prompt.responses.interactive.confirm_prompt_response import (
            ConfirmPromptResponse,
        )

        has_changes = self.has_changes_since_last_publication_tag()
        if not force and has_changes:
            self.log(f"Package {self.get_package_name()} has no new content to bump.")
            return False

        current_version = self.get_project_version()
        new_version = version_increment(version=current_version, **kwargs)
        branch_name = f"version-{new_version}"

        self.info(f"Bumping version to {new_version}", prefix=True)

        def _bump() -> None:
            from wexample_helpers_git.helpers.git import git_create_or_switch_branch

            # Create or switch to branch first, so changes are committed on it.
            git_create_or_switch_branch(
                branch_name, cwd=self.get_path(), inherit_stdio=True
            )
            self.log(message=f'Switched to branch "{branch_name}"', indentation=1)

            # Change version number on this branch
            self.write_config_value("global.version", new_version)

            self.log(
                message=f'Bumped from "{current_version}" to "{new_version}"',
                indentation=1,
            )

        if interactive:
            changes_message = (
                " The project contains changes since last publication."
                if has_changes
                else ""
            )

            confirm = self.confirm(
                f"Do you want to create a new version for package {self.get_package_name()} in @path{{{self.get_path()}}}?{changes_message} "
                f'This will create/switch to branch "{branch_name}".',
                choices=ConfirmPromptResponse.MAPPING_PRESET_YES_NO,
                default="yes",
            )

            if confirm.is_ok():
                _bump()
                return True
        else:
            _bump()
            return True
        return False

    def configure(self, config: DictConfig) -> None:
        super().configure(config=config)

        self._init_env(env_dict=self.get_env_parameters().to_dict())

    def count_source_code_lines(self) -> int:
        """Count total lines in source code files."""
        return self._count_code_lines(directories=self._get_source_code_directories())

    def count_source_files(self) -> int:
        """Count number of source code files."""
        return self._count_files(directories=self._get_source_code_directories())

    def count_test_code_lines(self) -> int:
        """Count total lines in test code files."""
        return self._count_code_lines(directories=self._get_test_code_directories())

    def count_test_files(self) -> int:
        """Count number of test code files."""
        return self._count_files(directories=self._get_test_code_directories())

    def ensure_app_manager(self) -> bool:
        from wexample_app.const.globals import APP_PATH_BIN_APP_MANAGER

        current_app_manager_path = self.get_path() / APP_PATH_BIN_APP_MANAGER

        if current_app_manager_path.exists():
            return True

        closest_app_manager_path = self.search_closest_app_manager_bin_path()
        if closest_app_manager_path != current_app_manager_path:
            self.log(f"Creating symlink: {current_app_manager_path}")

            # Remove if symlink already exists but point to a missing file.
            current_app_manager_path.unlink()
            current_app_manager_path.symlink_to(closest_app_manager_path.resolve())
            # False says newly created.
            return False
        return True

    def ensure_app_manager_setup(self) -> None:
        from wexample_app.const.globals import APP_PATH_BIN_APP_MANAGER

        self._override_pyproject_dependencies_by_current_distribution_versions()
        self.ensure_app_manager()

        self.shell_run_for_app(cmd=[str(APP_PATH_BIN_APP_MANAGER), "setup"])

    def get_app_env(self) -> str:
        from wexample_app.const.globals import ENV_VAR_NAME_APP_ENV

        return self.get_env_parameter(key=ENV_VAR_NAME_APP_ENV, default=ENV_NAME_PROD)

    def get_dependencies_versions(self) -> dict[str, str]:
        return {}

    def get_last_publication_tag(self) -> str | None:
        """Return the last publication tag for this package, or None if none exists."""
        from wexample_helpers_git.helpers.git import git_last_tag_for_prefix

        prefix = f"{self.get_package_name()}/v*"
        return git_last_tag_for_prefix(prefix, cwd=self.get_path(), inherit_stdio=False)

    def get_local_libraries_paths(self) -> list[ConfigValue]:
        return self.get_runtime_config().search(f"libraries").get_list_or_default()

    @abstract_method
    def get_package_import_name(self) -> str:
        pass

    @abstract_method
    def get_main_code_file_extension(self) -> str:
        pass

    def get_package_name(self) -> str:
        return self.get_project_name()

    # Publication helpers
    def get_publication_tag_name(self) -> str:
        """Return the conventional tag name for this package publication.

        Format: "{package_name}/v{version}"
        """
        return f"{self.get_package_name()}/v{self.get_project_version()}"

    def has_a_test(self) -> bool:
        test_dir = self.find_by_name(APP_PATH_TEST)
        return (
            test_dir
            and test_dir.is_directory()
            and any(test_dir.get_path().rglob("*.py"))
        )

    def has_changes_since_last_coverage(self) -> bool:
        """Return True if there are any changes (committed or not) since last coverage."""
        from wexample_helpers_git.helpers.git import git_has_uncommitted_changes

        last_commit = (
            self.get_config()
            .search("test.coverage.last_report.commit_hash")
            .get_str_or_default()
        )

        if not last_commit:
            return True

        return git_has_uncommitted_changes(
            cwd=self.get_path()
        ) or git_has_changes_since_tag(last_commit, cwd=self.get_path())

    def has_changes_since_last_publication_tag(self) -> bool:
        """Return True if there are changes in the package directory since the last publication tag.

        If there is no previous tag, returns True (first publication).
        """
        from wexample_helpers_git.helpers.git import git_has_changes_since_tag

        last_tag = self.get_last_publication_tag()
        if last_tag is None:
            return True
        # Limit diff to current package folder, run from package cwd using '.'
        return git_has_changes_since_tag(last_tag, ".", cwd=self.get_path())

    def libraries_sync(self) -> None:
        from wexample_wex_addon_app.commands.dependencies.publish import (
            app__dependencies__publish,
        )

        for library_path_config in (
            self.get_runtime_config().search("libraries").get_list_or_default()
        ):
            if library_path_config.is_str() and BasicAppWorkdir.is_app_workdir_path(
                path=library_path_config.get_str()
            ):
                publishable_dependencies = (
                    BasicAppWorkdir.manager_run_command_from_path(
                        command=app__dependencies__publish,
                        path=library_path_config.get_str(),
                    ).get_output()
                )

                self.update_dependencies(publishable_dependencies)
        self.io.success("All libraries versions are up to date.")

    def publish(self, force: bool = False) -> None:
        from wexample_helpers_git.const.common import GIT_BRANCH_MAIN

        if not self.should_be_published(force=force):
            return

        self._publish(force=force)
        self.success(
            f"Published {self.get_package_name()} as {self.get_publication_tag_name()}."
        )
        self.add_publication_tag()
        self.merge_to_main()
        self.push_to_deployment_remote(branch_name=GIT_BRANCH_MAIN)

    def publish_bumped(self, force: bool = False, interactive: bool = True) -> None:
        from wexample_wex_addon_app.commands.file_state.rectify import (
            app__file_state__rectify,
        )
        from wexample_wex_addon_app.commands.package.bump import app__package__bump
        from wexample_wex_addon_app.commands.package.commit_and_push import (
            app__package__commit_and_push,
        )
        from wexample_wex_addon_app.commands.package.publish import (
            app__package__publish,
        )
        from wexample_wex_addon_app.commands.version.propagate import (
            app__version__propagate,
        )

        if force or self.has_changes_since_last_publication_tag():
            # Reserve 1 unit on main progress bar, subdivided into 5 steps
            sub_progress = self.progress(
                total=5,
                color=TerminalColor.YELLOW,
                indentation=1,
                print_response=False,
            ).get_handle()

            sub_progress.advance(step=1, label=f"Bumping {self.get_project_name()}")
            bump_args = []
            if force:
                bump_args.append("--force")
            if not interactive:
                bump_args.append("--yes")
            bump_response = self.manager_run_command(
                command=app__package__bump, arguments=bump_args
            ).get_output_value()

            # Bump cancelled.
            if not bump_response.is_true():
                return

            sub_progress.advance(
                step=1, label=f"Rectifying file state for {self.get_project_name()}"
            )
            rectify_args = ["--loop"]
            if not interactive:
                rectify_args.append("--yes")
            self.manager_run_command(
                command=app__file_state__rectify, arguments=rectify_args
            )

            # TODO run tests

            sub_progress.advance(
                step=1, label=f"Committing and pushing {self.get_project_name()}"
            )
            self.manager_run_command(command=app__package__commit_and_push)

            sub_progress.advance(
                step=1, label=f"Propagating version for {self.get_project_name()}"
            )
            self.manager_run_command(command=app__version__propagate)

            sub_progress.advance(step=1, label=f"Publishing {self.get_project_name()}")
            self.manager_run_command(
                command=app__package__publish, arguments=(["--force"] if force else [])
            )
        else:
            self.io.log("No change to publish, skipping.")

    def publish_dependencies(self) -> dict[str, str]:
        """Publish witch dependency **current package represents for others**,
        not the dependencies the current package is dependent on.
        """
        return {self.get_package_name(): self.get_project_version()}

    def search_app_or_suite_runtime_config(
        self, key_path: str, default: Any = None
    ) -> ConfigValue:
        def _test_path(workdir) -> Path | None:
            config = workdir.get_runtime_config().search(path=key_path)
            if not config.is_none():
                return config
            return None

        return self.search_closest_in_suites_tree(callback=_test_path) or ConfigValue(
            raw=default
        )

    def search_closest_app_manager_bin_path(self) -> Path | None:
        def _test_path(workdir) -> Path | None:
            from wexample_app.const.globals import APP_PATH_BIN_APP_MANAGER

            manager_bin_path = workdir.get_path() / APP_PATH_BIN_APP_MANAGER

            if manager_bin_path.exists():
                return manager_bin_path

        return self.search_closest_in_suites_tree(callback=_test_path)

    def set_app_env(self, env: str | None) -> None:
        from wexample_app.const.globals import ENV_VAR_NAME_APP_ENV

        self.set_env_parameter(key=ENV_VAR_NAME_APP_ENV, value=env)

        self.get_registry(rebuild=True)

    def setup_install(self, env: str | None = None, force: bool = False) -> None:
        env = env or self.get_app_env()
        self.log(f"Set app environment to @🔵+bold{{{env}}}...")
        self.set_app_env(env=env)

        self.log("Check manager setup...")
        self.ensure_app_manager_setup()

        self.log("Install app...")
        self.app_install(env, force=force)

    def should_be_published(self, force: bool = False) -> bool:
        current_tag = self.get_publication_tag_name()
        last_tag = self.get_last_publication_tag()
        if not force and last_tag == current_tag:
            self.log(f"{self.get_package_name()} already published as {current_tag}.")
            return False
        return True

    def update_dependencies(self, dependencies_map: dict[str, str]) -> None:
        # Let language specific workdir manage how to update.
        pass

    def _count_code_lines(
        self,
        directories: list[TargetFileOrDirectoryType],
    ) -> int:
        """Count total lines in files matching the main code extension."""
        total = 0

        for item in directories:
            total += line_count_recursive(
                path=item.get_path(), pattern=f"*{self.get_main_code_file_extension()}"
            )

        return total

    def _count_files(
        self,
        directories: list[TargetFileOrDirectoryType],
    ) -> int:
        """Count files matching the main code extension in given directories."""
        total = 0
        extension = self.get_main_code_file_extension()

        for item in directories:
            path = item.get_path()
            if path.exists() and path.is_dir():
                # Count files with the main code extension
                total += len(list(path.rglob(f"*{extension}")))

        return total

    def _get_source_code_directories(self) -> list[TargetFileOrDirectoryType]:
        return []

    def _get_suite_package_workdir_class(self) -> type[FrameworkPackageSuiteWorkdir]:
        from wexample_wex_addon_app.workdir.framework_packages_suite_workdir import (
            FrameworkPackageSuiteWorkdir,
        )

        """ A suite can be a sub-suite"""
        return FrameworkPackageSuiteWorkdir

    def _get_test_code_directories(self) -> list[TargetFileOrDirectoryType]:
        return []

    def _override_pyproject_dependencies_by_current_distribution_versions(self) -> None:
        from pathlib import Path

        import tomlkit
        from wexample_helpers.helpers.module import module_get_distribution_map

        installed = module_get_distribution_map()

        manager_path = self.get_path() / APP_PATH_APP_MANAGER
        pyproject_path = Path(manager_path / "pyproject.toml")

        # Read raw text for tomlkit (to preserve comments)
        raw = pyproject_path.read_text()

        # Parse with tomlkit
        doc = tomlkit.parse(raw)

        # "project.dependencies" is a TOML array with preserved formatting
        deps = doc.get("project", {}).get("dependencies", [])

        updated = False
        new_deps = tomlkit.array().multiline(True)

        for dep in deps:
            dep_str = str(dep)

            # Parse dependency into (name, version)
            if "==" in dep_str:
                name, _ = dep_str.split("==", 1)
            elif ">=" in dep_str:
                name, _ = dep_str.split(">=", 1)
            elif "<=" in dep_str:
                name, _ = dep_str.split("<=", 1)
            else:
                # keep untouched (no version specified)
                new_deps.append(dep)
                continue

            pkg_name = name.strip().lower()
            installed_version = installed.get(pkg_name)

            if installed_version:
                new_dep = f"{pkg_name}=={installed_version}"
                if dep_str != new_dep:
                    self.log(f"{dep_str} → {new_dep}")
                    updated = True
                new_deps.append(new_dep)
            else:
                new_deps.append(dep)

        # If changed, replace the array in the document
        if updated:
            doc["project"]["dependencies"] = new_deps

            # Write back, preserving original formatting + comments
            pyproject_path.write_text(tomlkit.dumps(doc))

            self._python_export_dependencies(path=manager_path)

            self.log("✓ pyproject.toml updated")

    def _publish(self, force: bool = False) -> None:
        """Internal publication process"""

    def _python_export_dependencies(self, path: Path) -> None:
        import sys

        self.shell_run_from_path(
            cmd=[
                f"{sys.prefix}/bin/pip-compile",
                f"{path}/pyproject.toml",
                "--output-file",
                f"{path}/requirements.txt",
            ],
            path=path,
        )



from __future__ import annotations

from wexample_app.exception.app_runtime_exception import AppRuntimeException


class DependencyViolationException(AppRuntimeException):
    """Exception raised when a package imports code from another package without a declared dependency."""

    error_code: str = "DEPENDENCY_VIOLATION"

    def __init__(
        self,
        package_name: str,
        imported_package: str,
        import_locations: list[str],
        cause: Exception | None = None,
        previous: Exception | None = None,
    ) -> None:
        from wexample_helpers.helpers.cli import cli_make_clickable_path

        imports_details = "\n".join(
            f" - {cli_make_clickable_path(loc)}" for loc in import_locations
        )

        data = {
            "package_name": package_name,
            "imported_package": imported_package,
            "import_count": len(import_locations),
        }

        super().__init__(
            message=(
                f'Dependency violation: package "{package_name}" imports code from "{imported_package}" '
                f"but there is no declared local dependency path. "
                f'Add "{imported_package}" to the \'project.dependencies\' of "{package_name}" in its pyproject.toml, '
                f'or declare an intermediate local package that depends on "{imported_package}".\n\n'
                f"Detected import locations:\n{imports_details}"
            ),
            data=data,
            cause=cause,
            previous=previous,
        )



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any

from wexample_config.config_value.config_value import ConfigValue
from wexample_helpers.classes.base_class import BaseClass
from wexample_helpers.classes.field import public_field
from wexample_helpers.decorator.base_class import base_class

if TYPE_CHECKING:
    from wexample_wex_addon_app.workdir.framework_packages_suite_workdir import (
        FrameworkPackageSuiteWorkdir,
    )


@base_class
class WithSuiteTreeWorkdirMixin(BaseClass):
    _suite_workdir: None | False | FrameworkPackageSuiteWorkdir = public_field(
        default=None, description="Cache reference to the parent suite"
    )
    _suite_workdir_path: None | False | Path = public_field(
        default=None, description="Cache reference to the parent suite"
    )

    def collect_stack_in_suites_tree(
        self, callback: Callable[[Any], Any], include_self: bool = True
    ) -> list[Any]:
        """
        Walk the suite tree from the current workdir and collect callback results.
        Optionally ignore the current workdir ("self") as starting point.
        """
        workdir = self if include_self else self.get_shallow_suite_workdir()
        stack: list[Any] = []

        while workdir:
            result = callback(workdir)
            if result is not None:
                stack.append(result)

            workdir = workdir.get_shallow_suite_workdir()

        return stack

    def find_suite_workdir_path(self) -> Path | None:
        """
        We have to trust the configuration file to know if parent directory is a suite or not,
        as we cannot directly load suite python class from a different venv.
        """
        from wexample_helpers.helpers.directory import directory_iterate_parent_dirs

        from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir

        if self._suite_workdir_path is None:
            source_path = self.get_path()
            self._suite_workdir_path = False

            def _found(path: Path) -> bool:
                config = BasicAppWorkdir.get_config_from_path(
                    path=path,
                )
                if config:
                    if not config.read_config().search("package_suite").is_none():
                        return True

                return False

            suite_path = directory_iterate_parent_dirs(
                path=source_path.parent, condition=_found
            )

            if suite_path and suite_path != source_path:
                self._suite_workdir_path = suite_path

        return self._suite_workdir_path

    def get_env_parameter_or_suite_fallback(
        self, key: str, default: str | None = None
    ) -> str | None:
        value = self.get_env_parameter(
            key=key,
            default=default,
        )

        if value is None:
            suite_workdir = self.get_shallow_suite_workdir()
            if suite_workdir:
                return suite_workdir.get_env_parameter(
                    key=key,
                    default=default,
                )
        return value

    def get_shallow_suite_workdir(self) -> False | FrameworkPackageSuiteWorkdir:
        suite_path = self.find_suite_workdir_path()

        if suite_path and suite_path.exists():
            return self._get_suite_package_workdir_class().create_from_path(
                path=suite_path, configure=False
            )

    def get_suite_workdir(
        self, reload: bool = False
    ) -> False | FrameworkPackageSuiteWorkdir:
        if reload or self._suite_workdir is None:
            suite_path = self.find_suite_workdir_path()
            self._suite_workdir = False

            if suite_path and suite_path.exists():
                self._suite_workdir = (
                    self._get_suite_package_workdir_class().create_from_path(
                        path=suite_path
                    )
                )

        return self._suite_workdir

    def get_vendor_name(self) -> str:
        return self.search_in_package_or_suite_config(
            "global.vendor"
        ).get_str_or_default(default="acme")

    def propagate_version(self) -> None:
        self.get_suite_workdir().propagate_version_of(package=self)

    def save_dependency_from_package(
        self, package: FrameworkPackageSuiteWorkdir
    ) -> bool:
        """Add a dependency from another package, use strict version as this is the intended internal management."""
        return self.save_dependency(
            package_name=package.get_package_name(),
            version=package.get_project_version(),
        )

    def search_closest_in_suites_tree(self, callback) -> Any:
        workdir = self

        while workdir:
            result = callback(workdir)
            if result is not None:
                return result

            workdir = workdir.get_shallow_suite_workdir()

        return None

    def search_in_package_or_suite_config(self, key: str) -> ConfigValue:
        """Search for a config value in the package config, fallback to suite config if not found."""
        from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir

        value = self.get_config().search(key)

        if value.is_empty():
            suite_path = self.find_suite_workdir_path()
            if suite_path:
                # Also avoid using children tree as method may be executed before configuration process.
                suite_config_file = BasicAppWorkdir.get_config_from_path(
                    path=suite_path,
                )

                if suite_config_file:
                    return suite_config_file.read_config().search(key)

        return value



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_helpers.classes.base_class import BaseClass
from wexample_helpers.decorator.base_class import base_class
from wexample_wex_addon_app.config_value.app_license_config_value import (
    AppLicenseConfigValue,
)
from wexample_wex_addon_app.const.path import APP_PATH_LICENSE

if TYPE_CHECKING:
    from wexample_config.const.types import DictConfig


@base_class
class WithLicenseWorkdirMixin(BaseClass):
    def append_license(self, config: DictConfig | None = None) -> DictConfig:
        from wexample_filestate.const.disk import DiskItemType
        from wexample_filestate.option.text_option import TextOption

        config.get("children").append(
            {
                "name": APP_PATH_LICENSE,
                "type": DiskItemType.FILE,
                "should_exist": True,
                "content": self._get_license_content(),
                TextOption.get_name(): {"end_new_line": True},
            }
        )

        return config

    def _get_license_content(self) -> AppLicenseConfigValue:
        return AppLicenseConfigValue(workdir=self)



from __future__ import annotations

from tomlkit import TOMLDocument
from wexample_filestate.item.file.structured_content_file import StructuredContentFile
from wexample_helpers.classes.abstract_method import abstract_method
from wexample_helpers.classes.base_class import BaseClass
from wexample_helpers.decorator.base_class import base_class


@base_class
class WithAppConfigWorkdirMixin(BaseClass):
    def get_app_config(self, reload: bool = True) -> TOMLDocument:
        """
        Fetch the data from the pyproject.toml file.
        """
        return self.get_app_config_file(reload=reload).read_parsed()

    @abstract_method
    def get_app_config_file(self, reload: bool = True) -> StructuredContentFile:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware
from wexample_wex_core.decorator.option import option

from wexample_wex_addon_app.middleware.app_middleware import AppMiddleware
from wexample_wex_addon_app.middleware.each_suite_package_middleware import (
    EachSuitePackageMiddleware,
)
from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext


@option(name="yes", type=bool, default=False, is_flag=True)
@option(name="dry_run", type=bool, default=False, is_flag=True)
@option(name="loop", type=bool, default=False, is_flag=True)
@option(name="loop_limit", type=int, default=10)
@option(name="no_remote", type=bool, default=False, is_flag=True)
@option(name="filter_path", type=str, default=None)
@option(name="filter_operation", type=str, default=None)
@option(name="max", type=int, default=None)
@option(name="force", type=bool, default=False, is_flag=True)
@middleware(middleware=AppMiddleware)
@middleware(middleware=EachSuitePackageMiddleware)
@command(type=COMMAND_TYPE_ADDON)
def app__file_state__rectify(
    context: ExecutionContext,
    app_workdir: BasicAppWorkdir,
    force: bool = False,
    yes: bool = False,
    dry_run: bool = False,
    loop: bool = False,
    loop_limit: int = 10,
    no_remote: bool = False,
    filter_path: str | None = None,
    filter_operation: str | None = None,
    max: int = None,
) -> None:
    from wexample_filestate.enum.scopes import Scope

    if not dry_run:
        # Apply changes once, or keep looping until no operations remain (when --loop is set).
        iterations = 0
        while True:
            iterations += 1
            workdir = context.request.get_addon_manager().create_app_workdir()

            # Remove remote.
            scopes = (set(Scope) - {Scope.REMOTE}) if no_remote else None

            result = workdir.apply(
                interactive=(not yes),
                force=force,
                scopes=scopes,
                filter_path=filter_path,
                filter_operation=filter_operation,
                max=max,
            )

            if len(result.operations) == 0:
                pass_text = "pass" if iterations == 1 else "passes"
                context.io.success(
                    f"Rectification completed successfully after {iterations} {pass_text}."
                )
                break

            # Stop immediately after the first pass if looping is disabled.
            if not loop:
                operation_text = (
                    "operation" if len(result.operations) == 1 else "operations"
                )
                context.io.log(
                    f"Rectification pass completed; applied {len(result.operations)} {operation_text}."
                )
                break

            if iterations >= loop_limit:
                context.io.warning(
                    f"Loop limit reached ({iterations}/{loop_limit}); stopping further passes."
                )
                break

            context.io.log(
                f"Pass {iterations} completed with {len(result.operations)} operation(s); starting pass {iterations + 1} of {loop_limit}."
            )
    else:
        workdir = context.request.get_addon_manager().create_app_workdir(reload=True)
        workdir.dry_run()



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_helpers.classes.private_field import private_field
from wexample_helpers.decorator.base_class import base_class

from wexample_wex_addon_app.middleware.app_middleware import AppMiddleware
from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir

if TYPE_CHECKING:
    from wexample_wex_core.common.command_request import CommandRequest


@base_class
class PackageSuiteMiddleware(AppMiddleware):
    _fail_if_not_suite_workdir: bool = private_field(
        default=True, description="The called method works only on package suite"
    )

    def _create_app_workdir(
        self, request: CommandRequest, app_path: str
    ) -> BasicAppWorkdir:
        """Create and validate that the app workdir is a FrameworkPackageSuiteWorkdir."""
        from wexample_wex_addon_app.exception.invalid_workdir_type_exception import (
            InvalidWorkdirTypeException,
        )
        from wexample_wex_addon_app.workdir.framework_packages_suite_workdir import (
            FrameworkPackageSuiteWorkdir,
        )

        app_workdir = super()._create_app_workdir(request=request, app_path=app_path)

        # Validate that app_workdir is a FrameworkPackageSuiteWorkdir
        if self._fail_if_not_suite_workdir and not self._is_package_suite_workdir(
            workdir=app_workdir
        ):
            suite_path = app_workdir.find_suite_workdir_path()
            raise InvalidWorkdirTypeException(
                workdir_path=app_workdir.get_path(),
                actual_type=app_workdir.__class__.__name__,
                expected_type=FrameworkPackageSuiteWorkdir.__name__,
                suite_path=suite_path,
            )

        return app_workdir

    def _is_package_suite_workdir(self, workdir: BasicAppWorkdir) -> bool:
        from wexample_wex_addon_app.workdir.framework_packages_suite_workdir import (
            FrameworkPackageSuiteWorkdir,
        )

        return isinstance(workdir, FrameworkPackageSuiteWorkdir)



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware

from wexample_wex_addon_app.middleware.package_suite_middleware import (
    PackageSuiteMiddleware,
)
from wexample_wex_addon_app.workdir.framework_packages_suite_workdir import (
    FrameworkPackageSuiteWorkdir,
)

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext


@middleware(middleware=PackageSuiteMiddleware)
@command(
    type=COMMAND_TYPE_ADDON,
    description="Validate internal dependencies and propagate versions across all packages in the suite.",
)
def app__dependencies__check(
    context: ExecutionContext,
    app_workdir: FrameworkPackageSuiteWorkdir,
) -> None:
    # Validate internal dependencies
    app_workdir.packages_validate_internal_dependencies_declarations()



from __future__ import annotations

from wexample_app.exception.app_runtime_exception import AppRuntimeException


class GitRemoteException(AppRuntimeException):
    """Exception raised when a git remote operation fails."""

    error_code: str = "GIT_REMOTE_ERROR"

    def __init__(
        self,
        workdir_path: str,
        package_name: str,
        operation: str = "push",
        remote_name: str = "origin",
        branch_name: str | None = None,
        cause: Exception | None = None,
        previous: Exception | None = None,
    ) -> None:
        from wexample_helpers.helpers.cli import cli_make_clickable_path

        clickable_path = cli_make_clickable_path(workdir_path)

        data = {
            "workdir_path": workdir_path,
            "package_name": package_name,
            "operation": operation,
            "remote_name": remote_name,
        }

        message = (
            f"Git remote operation '{operation}' failed for package {package_name} at {clickable_path}. "
            f"The remote '{remote_name}' does not appear to be configured correctly."
        )

        if branch_name:
            message += f"\nBranch: {branch_name}"
            data["branch_name"] = branch_name

        message += (
            f"\n\nPossible causes:"
            f"\n  - The remote '{remote_name}' is not configured"
            f"\n  - The remote URL is invalid or inaccessible"
            f"\n  - The upstream branch does not exist on the remote"
            f"\n\nTo fix this, you may need to:"
            f"\n  - Configure the remote: git remote add {remote_name} <url>"
            f"\n  - Set upstream: git push -u {remote_name} <branch>"
        )

        super().__init__(
            message=message,
            data=data,
            cause=cause,
            previous=previous,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware
from wexample_wex_core.decorator.option import option

from wexample_wex_addon_app.middleware.package_suite_middleware import (
    PackageSuiteMiddleware,
)
from wexample_wex_addon_app.workdir.framework_packages_suite_workdir import (
    FrameworkPackageSuiteWorkdir,
)

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext


@option(
    name="command",
    type=str,
    required=True,
    description='The full shell command to execute, e.g. "ls -la"',
)
@middleware(middleware=PackageSuiteMiddleware)
@command(
    type=COMMAND_TYPE_ADDON,
    description="Bump version for one or all package of the suite.",
)
def app__suite__exec_shell(
    context: ExecutionContext,
    command: str,
    app_workdir: FrameworkPackageSuiteWorkdir,
) -> None:
    from wexample_helpers.helpers.shell import shell_split_cmd

    app_workdir.packages_execute_shell(cmd=shell_split_cmd(command))



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware
from wexample_wex_core.decorator.option import option

from wexample_wex_addon_app.middleware.app_middleware import AppMiddleware

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext

    from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir


@option(
    name="env",
    type=str,
    default=None,
    description="Indicate to use environment-specific installation (e.g. local editable packages)",
)
@option(
    name="force",
    type=bool,
    default=False,
    description="Force reinstallation of all packages",
)
@middleware(middleware=AppMiddleware)
@command(
    type=COMMAND_TYPE_ADDON,
    description="Install package manager into suite or packages",
)
def app__setup__install(
    context: ExecutionContext,
    app_workdir: BasicAppWorkdir,
    env: str | None = None,
    force: bool = False,
) -> None:
    app_workdir.setup_install(env=env, force=force)



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_app.workdir.mixin.with_version_workdir_mixin import (
    WithVersionWorkdirMixin,
)
from wexample_filestate.testing.abstract_workdir_mixin_test import (
    AbstractWorkdirMixinTest,
)
from wexample_helpers.classes.base_class import BaseClass
from wexample_helpers.decorator.base_class import base_class

if TYPE_CHECKING:
    from wexample_config.const.types import DictConfig


class TestWithVersionWorkdirMixin(AbstractWorkdirMixinTest):
    """Test WithVersionWorkdirMixin functionality."""

    def _apply_mixin_to_config(self, mixin_instance, config: DictConfig) -> DictConfig:
        """Apply the version mixin method to enhance the config."""
        return mixin_instance.append_version(config)

    def _assert_applied(self, tmp_path) -> None:
        # Check that version.txt exists
        version_file = tmp_path / "version.txt"
        assert version_file.exists(), "version.txt should be created by the mixin"

    def _assert_not_applied(self, tmp_path) -> None:
        # Check that version.txt exists
        version_file = tmp_path / "version.txt"
        assert not version_file.exists(), "version.txt should be created by the mixin"

    def _get_apply_count(self) -> int:
        """Version mixin needs 3 applies: 1 for file creation, 1 for content writing, 1 for text processing."""
        return 2

    def _get_expected_files(self) -> list[str]:
        """Return list of files that should be created by the version mixin."""
        return ["version.txt"]

    def _get_mixin_config(self) -> DictConfig:
        """Return the base configuration for the version mixin test."""
        return {"children": []}

    def _get_test_workdir_class(self) -> type:
        """Return the test class that inherits from WithVersionWorkdirMixin."""

        @base_class
        class VersionWorkdir(WithVersionWorkdirMixin, BaseClass):
            """Test class that inherits from WithVersionWorkdirMixin."""

        return VersionWorkdir



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_filestate.testing.abstract_workdir_mixin_test import (
    AbstractWorkdirMixinTest,
)
from wexample_helpers.classes.base_class import BaseClass
from wexample_helpers.decorator.base_class import base_class

from wexample_wex_addon_app.workdir.mixin.with_readme_workdir_mixin import (
    WithReadmeWorkdirMixin,
)

if TYPE_CHECKING:
    from wexample_config.const.types import DictConfig


class TestWithReadmeWorkdirMixin(AbstractWorkdirMixinTest):
    """Test WithReadmeWorkdirMixin functionality."""

    def test_readme_file_created(self, tmp_path) -> None:
        """Test that README.md is created by the mixin."""
        self._setup_with_tmp_path(tmp_path)

        # Create workdir manager with readme mixin
        manager = self._create_test_workdir_manager(tmp_path)

        # Apply once to create the file
        manager.apply()

        # Check that README.md exists
        readme_file = tmp_path / "README.md"
        assert readme_file.exists(), "README.md should be created by the mixin"

    def _apply_mixin_to_config(self, mixin_instance, config: DictConfig) -> DictConfig:
        """Apply the readme mixin method to enhance the config."""
        return mixin_instance.append_readme(config)

    def _assert_applied(self, tmp_path) -> None:
        """Assert that README.md exists."""
        readme_file = tmp_path / "README.md"
        assert readme_file.exists(), "README.md should be created by the mixin"

    def _assert_not_applied(self, tmp_path) -> None:
        """Assert that README.md does not exist."""
        readme_file = tmp_path / "README.md"
        assert not readme_file.exists(), "README.md should not exist"

    def _get_apply_count(self) -> int:
        """Readme mixin needs 2 applies: 1 for file creation, 1 for content writing."""
        return 2

    def _get_expected_files(self) -> list[str]:
        """Return list of files that should be created by the readme mixin."""
        return ["README.md"]

    def _get_mixin_config(self) -> DictConfig:
        """Return the base configuration for the readme mixin test."""
        return {"children": []}

    def _get_test_workdir_class(self) -> type:
        """Return the test class that inherits from WithReadmeWorkdirMixin."""

        @base_class
        class ReadmeWorkdir(WithReadmeWorkdirMixin, BaseClass):
            """Test class that inherits from WithReadmeWorkdirMixin."""

        return ReadmeWorkdir



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_helpers.decorator.base_class import base_class
from wexample_wex_core.middleware.abstract_middleware import AbstractMiddleware

if TYPE_CHECKING:
    from wexample_app.command.option import Option
    from wexample_helpers.const.types import Kwargs
    from wexample_wex_core.common.command_method_wrapper import CommandMethodWrapper
    from wexample_wex_core.common.command_request import CommandRequest
    from wexample_wex_core.context.execution_context import ExecutionContext


@base_class
class AppMiddleware(AbstractMiddleware):
    def build_execution_contexts(
        self,
        command_wrapper: CommandMethodWrapper,
        request: CommandRequest,
        function_kwargs: Kwargs,
    ) -> list[ExecutionContext]:
        app_path = function_kwargs.get(
            "app_path", str(request.kernel.call_workdir.get_path())
        )

        function_kwargs.pop("app_path", None)
        function_kwargs["app_workdir"] = self._create_app_workdir(
            request=request, app_path=app_path
        )

        return super().build_execution_contexts(
            command_wrapper=command_wrapper,
            request=request,
            function_kwargs=function_kwargs,
        )

    def _create_app_workdir(self, request: CommandRequest, app_path: str):
        """Create and return the app workdir. Can be overridden by subclasses to add validation."""
        return request.get_addon_manager().create_app_workdir(path=app_path)

    def _get_middleware_options(self) -> list[Option]:
        """Get the default file option definition."""
        from wexample_app.command.option import Option

        return [
            Option(
                name="app_path",
                type=str,
                required=False,
                description="Path to the app directory (defaults to current directory)",
            )
        ]



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware
from wexample_wex_core.decorator.option import option

from wexample_wex_addon_app.middleware.package_suite_middleware import (
    PackageSuiteMiddleware,
)
from wexample_wex_addon_app.workdir.framework_packages_suite_workdir import (
    FrameworkPackageSuiteWorkdir,
)

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext


@option(name="yes", type=bool, default=False, is_flag=True)
@option(name="force", type=bool, default=False, is_flag=True)
@option(name="ignore_dependencies", type=bool, default=False, is_flag=True)
@middleware(middleware=PackageSuiteMiddleware)
@command(
    type=COMMAND_TYPE_ADDON,
    description="Publish package to package manager (npm, PyPI, packagist, etc.). Use --all-packages to publish all packages in suite.",
)
def app__suite__publish(
    context: ExecutionContext,
    app_workdir: FrameworkPackageSuiteWorkdir,
    yes: bool = False,
    force: bool = False,
    ignore_dependencies: bool = False,
) -> None:
    from wexample_prompt.enums.terminal_color import TerminalColor

    if not ignore_dependencies:
        app_workdir.packages_validate_internal_dependencies_declarations()

    packages = app_workdir.get_ordered_packages()

    context.io.log("Starting deployment...")
    context.io.indentation_up()
    progress = context.io.progress(
        total=len(packages),
        print_response=False,
        color=TerminalColor.CYAN,
    ).get_handle()

    # Process packages in order leaf -> trunk.
    # Use manager for every command allow to use complete specific environment.
    for package in packages:
        progress.advance(label=f"Publishing {package.get_project_name()}", step=1)
        package.publish_bumped(force=force, interactive=not yes)

    progress.finish(label="All packages published successfully")



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

from wexample_helpers.const.types import StringKeysDict

if TYPE_CHECKING:
    from wexample_helpers.const.types import StringKeysDict


class AppConfig(TypedDict):
    branch: str | None
    domain_main: str
    domain_tld: str
    domains: list[str]
    domains_string: str
    env: StringKeysDict
    host: dict[str, str]
    name: str
    password: dict[str, str]
    path: dict[str, str]
    server: StringKeysDict
    service: StringKeysDict
    started: bool
    user: dict[str, str | int]



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_helpers_git.const.common import GIT_BRANCH_MAIN, GIT_REMOTE_ORIGIN
from wexample_helpers_git.helpers.git import git_run
from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir

if TYPE_CHECKING:
    from wexample_config.options_provider.abstract_options_provider import (
        AbstractOptionsProvider,
    )
    from wexample_prompt.common.progress.progress_handle import ProgressHandle


class CodeBaseWorkdir(BasicAppWorkdir):
    def add_publication_tag(self) -> None:
        from wexample_helpers_git.helpers.git import (
            git_push_tag,
            git_tag_annotated,
            git_tag_exists,
        )

        cwd = self.get_path()
        tag = f"{self.get_package_name()}/v{self.get_project_version()}"

        # Create the annotated tag if it does not already exist locally.
        if not git_tag_exists(tag, cwd=cwd, inherit_stdio=False):
            git_tag_annotated(tag, f"Release {tag}", cwd=cwd, inherit_stdio=True)
        else:
            self.warning(f"Tag {tag} already exists locally; pushing it.")

        # Push the tag explicitly to the remote to ensure it's published.
        git_push_tag(tag, cwd=cwd, inherit_stdio=True)

    def _build_dependency_string(self, package_name: str, version: str) -> str:
        return f"{package_name}=={version}"

    def build_dependencies_stack(
        self, package: CodeBaseWorkdir, dependency: CodeBaseWorkdir
    ) -> list[CodeBaseWorkdir]:
        """When package is dependent from another one (is using it in its codebase),
        list the packages inheritance stack to find the original package declaring the explicit dependency
        """
        return []

    def commit_changes(
        self,
        progress: ProgressHandle | None = None,
    ) -> None:
        """Commit local changes (if any), without pushing."""
        from wexample_helpers_git.helpers.git import (
            git_commit_all_with_message,
            git_ensure_upstream,
            git_has_index_changes,
            git_has_working_changes,
            git_pull_rebase_autostash,
        )

        cwd = self.get_path()
        progress = (
            progress
            or self.progress(label="Committing changes...", total=3).get_handle()
        )

        git_ensure_upstream(
            cwd=cwd,
            default_remote=self._get_deployment_remote_name(),
            inherit_stdio=True,
        )
        progress.advance(step=1, label="Ensured upstream")

        git_pull_rebase_autostash(cwd=cwd, inherit_stdio=True)
        progress.advance(step=1, label="Pulled latest (rebase)")

        has_working_changes = git_has_working_changes(cwd=cwd)
        has_index_changes = git_has_index_changes(cwd=cwd)

        if has_working_changes or has_index_changes:
            git_commit_all_with_message(
                f"Publishing version {self.get_project_version()}",
                cwd=cwd,
                inherit_stdio=True,
            )
            progress.finish(label="Committed changes")
        else:
            progress.finish(label="No changes to commit")

    def depends_from(self, package: CodeBaseWorkdir) -> bool:
        for dependence_name in self.get_dependencies_versions().keys():
            if package.get_package_name() == dependence_name:
                return True
        return False

    def get_io_context_prefix(self) -> str | None:
        from wexample_helpers.helpers.cli import cli_make_clickable_path

        """Get the prefix to prepend to messages (e.g., '[child]')."""
        return cli_make_clickable_path(self.get_path(), self.get_project_name())

    def get_io_context_prefix_format(self) -> str:
        return "‹› {prefix} | "

    def get_options_providers(self) -> list[type[AbstractOptionsProvider]]:
        from wexample_filestate.options_provider.default_options_provider import (
            DefaultOptionsProvider,
        )
        from wexample_filestate_git.options_provider.git_options_provider import (
            GitOptionsProvider,
        )

        return [
            DefaultOptionsProvider,
            GitOptionsProvider,
        ]

    def has_working_changes(self) -> bool:
        from wexample_helpers_git.helpers.git import git_has_working_changes

        return git_has_working_changes(cwd=self.get_path(), inherit_stdio=True)

    def imports_package_in_codebase(self, searched_package: CodeBaseWorkdir) -> bool:
        """Check whether the given package is used in this package's codebase."""
        return False

    def merge_to_main(self, branch_name: str = GIT_BRANCH_MAIN) -> None:
        """Merge current branch into the specified main branch, then return to the original branch.

        :param branch_name: Name of the main branch (default: "main").

        This method:
        1. Saves the current branch name
        2. Merges main into the current branch (to ensure compatibility)
        3. Switches to the main branch
        4. Merges the current branch into main
        5. Returns to the original branch

        Raises if there are uncommitted changes or merge conflicts.
        """
        from wexample_helpers.helpers.shell import shell_run
        from wexample_helpers_git.helpers.git import (
            git_current_branch,
            git_has_uncommitted_changes,
            git_switch_branch,
        )

        cwd = self.get_path()

        # Ensure no uncommitted changes before starting
        if git_has_uncommitted_changes(cwd=cwd):
            raise RuntimeError(
                f"Cannot merge to {branch_name}: uncommitted changes detected. "
                "Please commit or stash your changes first."
            )

        # Save current branch name
        current_branch = git_current_branch(cwd=cwd, inherit_stdio=False)

        if current_branch == branch_name:
            self.warning(f"Already on {branch_name} branch, nothing to merge.")
            return

        try:
            # Step 1: Merge main into current branch to ensure compatibility
            self.info(f"Merging {branch_name} into {current_branch}...")
            shell_run(
                [
                    "git",
                    "merge",
                    branch_name,
                    "--no-ff",
                    "-m",
                    f"Merge branch '{branch_name}' into {current_branch}",
                ],
                inherit_stdio=True,
                cwd=cwd,
            )

            # Step 2: Switch to main
            self.info(f"Switching to {branch_name}...")
            git_switch_branch(branch_name, cwd=cwd, inherit_stdio=True)

            # Step 3: Merge current branch into main
            self.info(f"Merging {current_branch} into {branch_name}...")
            shell_run(
                [
                    "git",
                    "merge",
                    current_branch,
                    "--no-ff",
                    "-m",
                    f"Merge branch '{current_branch}' into {branch_name}",
                ],
                inherit_stdio=True,
                cwd=cwd,
            )

            self.success(f"Successfully merged {current_branch} into {branch_name}")

        finally:
            # Step 4: Always return to the original branch
            self.info(f"Returning to {current_branch}...")
            git_switch_branch(current_branch, cwd=cwd, inherit_stdio=True)

    def _get_deployment_remote_name(self) -> str | None:
        return self.search_app_or_suite_runtime_config(
            "git.main_deployment_remote_name", default=None
        ).get_str_or_none()

    def push_to_deployment_remote(self, branch_name: str | None = None) -> None:
        self.push_changes(
            remote_name=self._get_deployment_remote_name(),
            branch_name=branch_name,
        )

    def push_changes(
        self,
        remote_name: str | None = None,
        branch_name: str | None = None,
    ) -> None:
        remote = remote_name or GIT_REMOTE_ORIGIN
        branch_name = branch_name or GIT_BRANCH_MAIN

        local_branch, remote_branch = (
            branch_name.split(":", 1)
            if ":" in branch_name
            else (branch_name, branch_name)
        )

        self.git_run(
            cmd=[
                "push",
                remote,
                f"{local_branch}:{remote_branch}",
                "--follow-tags",
                "--force",
                "--porcelain",
            ],
            inherit_stdio=False,
        )

    def git_run(self, *args, **kwargs):
        return git_run(
            cwd=self.get_path(),
            *args,
            **kwargs,
        )

    def save_dependency(self, package_name: str, version: str) -> bool:
        """Add or update a dependency with strict version."""
        config = self.get_app_config_file()
        return config.add_dependency(package_name=package_name, version=version)



from __future__ import annotations

import shutil
from pathlib import Path

from wexample_helpers.const.types import PathOrString
from wexample_helpers.helpers.shell import shell_run


def python_ensure_pip_or_fail(venv_path: Path) -> bool:
    # Ensure pip is installed in the venv
    shell_run(
        cmd=[
            f"{venv_path}/bin/python",
            "-m",
            "ensurepip",
            "--upgrade",
        ],
        cwd=venv_path.parent,
        inherit_stdio=True,
    )

    return True


def python_install_dependencies_in_venv(
    venv_path: Path, names: list[PathOrString], editable: bool = False
) -> None:
    for dependency_name in names:
        python_install_dependency_in_venv(
            venv_path=venv_path, name=dependency_name, editable=editable
        )


def python_install_dependency_in_venv(
    venv_path: Path, name: PathOrString, editable: bool = False
) -> None:
    from wexample_helpers.helpers.shell import shell_run

    cmd = [
        f"{venv_path}/bin/python",
        "-m",
        "pip",
        "install",
    ]

    if editable:
        cmd.append("-e")

    cmd.append(str(name))

    shell_run(
        cmd=cmd,
        cwd=venv_path.parent,
        inherit_stdio=True,
    )


def python_install_environment(path: PathOrString) -> Path:
    project_path = Path(path)

    venv_path = project_path / ".venv"
    req_path = project_path / "requirements.txt"

    if venv_path.exists():
        shutil.rmtree(venv_path, ignore_errors=True)

    # Crée un nouveau venv
    shell_run(
        cmd=["python3", "-m", "venv", ".venv", "--clear", "--copies"],
        cwd=project_path,
        inherit_stdio=True,
    )

    if req_path.exists():
        python_bin = venv_path / "bin" / "python"

        shell_run(
            cmd=[
                str(python_bin),
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
                "setuptools",
                "wheel",
            ],
            cwd=project_path,
            inherit_stdio=True,
        )

        shell_run(
            cmd=[str(python_bin), "-m", "pip", "install", "-r", "requirements.txt"],
            cwd=project_path,
            inherit_stdio=True,
        )

    return venv_path


def python_is_package_installed_editable_in_venv(
    venv_path: Path,
    package_name: str,
    package_path,
) -> bool:
    """Return True if the package is installed in editable mode at the given path."""
    import subprocess
    from pathlib import Path

    python_bin = venv_path / "bin" / "python"

    try:
        result = subprocess.run(
            [str(python_bin), "-m", "pip", "show", package_name],
            capture_output=True,
            text=True,
            timeout=5,
        )

        if result.returncode != 0:
            return False

        editable_location = None

        for line in result.stdout.splitlines():
            if line.startswith("Editable project location:"):
                editable_location = line.split(":", 1)[1].strip()
                break

        if not editable_location:
            return False

        return Path(editable_location).resolve() == Path(package_path).resolve()

    except (subprocess.SubprocessError, OSError):
        return False



from __future__ import annotations

from pathlib import Path

# filestate: python-constant-sort
APP_PATH_EXAMPLES: Path = Path("examples")
APP_PATH_LICENSE: Path = Path("LICENSE")
APP_PATH_README: Path = Path("README.md")
APP_PATH_TEST: Path = Path("tests")



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware

from wexample_wex_addon_app.middleware.app_middleware import AppMiddleware
from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext


@middleware(middleware=AppMiddleware)
@command(type=COMMAND_TYPE_ADDON)
def app__libraries__sync(
    context: ExecutionContext,
    app_workdir: BasicAppWorkdir,
) -> None:
    app_workdir.libraries_sync()



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_helpers.classes.private_field import private_field
from wexample_helpers.decorator.base_class import base_class

from wexample_wex_addon_app.middleware.package_suite_middleware import (
    PackageSuiteMiddleware,
)

if TYPE_CHECKING:
    from wexample_app.command.option import Option
    from wexample_helpers.const.types import Kwargs
    from wexample_wex_core.common.command_method_wrapper import CommandMethodWrapper
    from wexample_wex_core.common.command_request import CommandRequest
    from wexample_wex_core.context.execution_context import ExecutionContext


@base_class
class SuiteOrEachPackageMiddleware(PackageSuiteMiddleware):
    """Middleware that allows executing on:
    - The suite itself (default when app_path points to a suite)
    - All packages (with --all-packages)
    - Only packages, excluding suite (with --all-packages --packages-only)
    - Only suite, excluding packages (with --suite-only)
    """

    _fail_if_not_suite_workdir: bool = private_field(
        default=False, description="Can be called both on suite or package"
    )

    def build_execution_contexts(
        self,
        command_wrapper: CommandMethodWrapper,
        request: CommandRequest,
        function_kwargs: Kwargs,
    ) -> list[ExecutionContext]:
        from wexample_wex_core.context.execution_context import ExecutionContext

        # Get flags
        all_packages = function_kwargs.get("all_packages", False)
        packages_only = function_kwargs.get("packages_only", False)
        suite_only = function_kwargs.get("suite_only", False)

        # Remove flags from kwargs
        function_kwargs.pop("all_packages", None)
        function_kwargs.pop("packages_only", None)
        function_kwargs.pop("suite_only", None)

        # Validate conflicting options
        if packages_only and suite_only:
            raise ValueError(
                "Cannot use both --packages-only and --suite-only at the same time."
            )

        # Call parent to get and validate app_workdir
        # Don't fail if not suite when we're not targeting packages
        original_fail_setting = self._fail_if_not_suite_workdir
        if not all_packages and not packages_only:
            self._fail_if_not_suite_workdir = False

        try:
            super().build_execution_contexts(
                command_wrapper=command_wrapper,
                request=request,
                function_kwargs=function_kwargs,
            )
        finally:
            self._fail_if_not_suite_workdir = original_fail_setting

        app_workdir = function_kwargs.get("app_workdir")
        is_suite = self._is_package_suite_workdir(workdir=app_workdir)

        contexts = []

        # Determine what to execute
        execute_suite = not packages_only and (suite_only or not all_packages)
        execute_packages = all_packages or packages_only

        # Execute on suite if requested
        if execute_suite:
            contexts.append(
                ExecutionContext(
                    middleware=self,
                    command_wrapper=command_wrapper,
                    request=request,
                    function_kwargs=function_kwargs.copy(),
                )
            )

        # Execute on packages if requested and we have a suite
        if execute_packages and is_suite:

            def custom_function(context: ExecutionContext, **kwargs) -> None:
                suite_workdir = kwargs.get("app_workdir")
                suite_workdir.packages_execute_manager(
                    command=request.resolver.build_command_from_function(
                        command_wrapper=command_wrapper
                    ),
                    arguments=request.arguments,
                )
                return None

            # Create a context with the custom function for package iteration
            package_kwargs = function_kwargs.copy()
            contexts.append(
                ExecutionContext(
                    middleware=self,
                    command_wrapper=command_wrapper,
                    request=request,
                    function_kwargs=package_kwargs,
                    function=custom_function,
                )
            )

        return contexts

    def _get_middleware_options(self) -> list[Option]:
        """Add options for controlling suite/package execution."""
        from wexample_app.command.option import Option

        options = super()._get_middleware_options()
        options.extend(
            [
                Option(
                    name="all_packages",
                    type=bool,
                    required=False,
                    default=False,
                    is_flag=True,
                    description="Execute the command on all packages of the suite (in addition to suite itself unless --packages-only is used)",
                ),
                Option(
                    name="packages_only",
                    type=bool,
                    required=False,
                    default=False,
                    is_flag=True,
                    description="Execute only on packages, not on the suite itself (implies --all-packages)",
                ),
                Option(
                    name="suite_only",
                    type=bool,
                    required=False,
                    default=False,
                    is_flag=True,
                    description="Execute only on the suite itself, not on packages",
                ),
            ]
        )
        return options



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware
from wexample_wex_core.decorator.option import option

from wexample_wex_addon_app.middleware.app_middleware import AppMiddleware

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext


@middleware(middleware=AppMiddleware)
@option(
    name="format",
    type=str,
    default="html",
    description="The output format of the report",
)
@command(
    type=COMMAND_TYPE_ADDON,
    description="Bump version for one or all package of the suite.",
)
def app__test__run(
    context: ExecutionContext, app_workdir: AppMiddleware, format: str | None = None
) -> None:
    app_workdir.test_run(format=format)



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_helpers.classes.private_field import private_field
from wexample_helpers.decorator.base_class import base_class

from wexample_wex_addon_app.middleware.package_suite_middleware import (
    PackageSuiteMiddleware,
)

if TYPE_CHECKING:
    from wexample_app.command.option import Option
    from wexample_helpers.const.types import Kwargs
    from wexample_wex_core.common.command_method_wrapper import CommandMethodWrapper
    from wexample_wex_core.common.command_request import CommandRequest
    from wexample_wex_core.context.execution_context import ExecutionContext


@base_class
class EachSuitePackageMiddleware(PackageSuiteMiddleware):
    _fail_if_not_suite_workdir: bool = private_field(
        default=False,
        description="Do not restrict the command to work only on package suite",
    )

    def build_execution_contexts(
        self,
        command_wrapper: CommandMethodWrapper,
        request: CommandRequest,
        function_kwargs: Kwargs,
    ) -> list[ExecutionContext]:
        from wexample_wex_core.context.execution_context import ExecutionContext

        # Check if --all-packages flag is set
        all_packages = function_kwargs.get("all_packages", False)
        function_kwargs.pop("all_packages", None)

        # Call parent to get and validate app_workdir,
        # but don't use generated contexts.
        super().build_execution_contexts(
            command_wrapper=command_wrapper,
            request=request,
            function_kwargs=function_kwargs,
        )

        if all_packages:
            suite_workdir = function_kwargs.get("app_workdir")

            if self._is_package_suite_workdir(workdir=suite_workdir):
                # Custom behavior: replace function with our custom one
                def custom_function(context: ExecutionContext, **kwargs) -> None:
                    suite_workdir.packages_execute_manager(
                        command=request.resolver.build_command_from_function(
                            command_wrapper=command_wrapper
                        ),
                        arguments=request.arguments,
                    )
                    return None

                # Create a single context with the custom function
                context = ExecutionContext(
                    middleware=self,
                    command_wrapper=command_wrapper,
                    request=request,
                    function_kwargs=function_kwargs,
                    function=custom_function,
                )

                return [context]

        # Nothing to execute.
        return []

    def _get_middleware_options(self) -> list[Option]:
        """Add the all_packages option."""
        from wexample_app.command.option import Option

        options = super()._get_middleware_options()
        options.append(
            Option(
                name="all_packages",
                type=bool,
                required=False,
                default=False,
                is_flag=True,
                description="Execute the command on all packages of the suite",
            )
        )
        return options



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware
from wexample_wex_core.decorator.option import option

from wexample_wex_addon_app.middleware.suite_or_each_package_middleware import (
    SuiteOrEachPackageMiddleware,
)
from wexample_wex_addon_app.workdir.code_base_workdir import CodeBaseWorkdir

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext


@middleware(middleware=SuiteOrEachPackageMiddleware)
@option(name="yes", type=bool, default=False, is_flag=True)
@option(name="force", type=bool, default=False, is_flag=True)
@command(
    type=COMMAND_TYPE_ADDON,
    description="Publish package to package manager (npm, PyPI, packagist, etc.). Use --all-packages to publish all packages in suite.",
)
def app__package__publish_bumped(
    context: ExecutionContext,
    app_workdir: CodeBaseWorkdir,
    yes: bool = False,
    force: bool = False,
) -> None:
    app_workdir.publish_bumped(
        interactive=not yes,
        force=force,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_helpers.classes.base_class import BaseClass
from wexample_helpers.decorator.base_class import base_class

from wexample_wex_addon_app.const.path import APP_PATH_README

if TYPE_CHECKING:
    from wexample_config.const.types import DictConfig
    from wexample_filestate.config_value.readme_content_config_value import (
        ReadmeContentConfigValue,
    )


@base_class
class WithReadmeWorkdirMixin(BaseClass):
    def append_readme(self, config: DictConfig | None = None) -> DictConfig:
        from wexample_filestate.const.disk import DiskItemType
        from wexample_filestate.option.text_option import TextOption

        config.get("children").append(
            {
                "name": APP_PATH_README,
                "type": DiskItemType.FILE,
                "should_exist": True,
                "content": self._get_readme_content(),
                TextOption.get_name(): {"end_new_line": True},
            }
        )

        return config

    def has_readme(self) -> bool:
        readme = self.find_by_name(APP_PATH_README)
        return readme.get_path().exists() if readme else None

    def _get_readme_content(self) -> ReadmeContentConfigValue:
        from wexample_wex_addon_app.config_value.app_readme_config_value import (
            AppReadmeConfigValue,
        )

        return AppReadmeConfigValue(workdir=self)



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware

from wexample_wex_addon_app.middleware.suite_or_each_package_middleware import (
    SuiteOrEachPackageMiddleware,
)
from wexample_wex_addon_app.workdir.code_base_workdir import CodeBaseWorkdir

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext


@middleware(middleware=SuiteOrEachPackageMiddleware)
@command(
    type=COMMAND_TYPE_ADDON,
    description="Commit and push changes for a package. Use --all-packages to apply to all packages in suite.",
)
def app__package__commit_and_push(
    context: ExecutionContext,
    app_workdir: CodeBaseWorkdir,
) -> None:
    from wexample_helpers_git.const.common import GIT_BRANCH_MAIN

    package_name = app_workdir.get_package_name()
    app_workdir.commit_changes()
    app_workdir.push_to_deployment_remote(branch_name=GIT_BRANCH_MAIN)
    context.io.success(f"Pushed {package_name}.")



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_helpers.classes.base_class import BaseClass
from wexample_helpers.const.types import FileStringOrPath
from wexample_helpers.decorator.base_class import base_class
from wexample_prompt.common.io_manager import IoManager

if TYPE_CHECKING:
    from wexample_config.config_value.nested_config_value import NestedConfigValue
    from wexample_filestate.item.file.yaml_file import YamlFile


@base_class
class WithAppRegistryMixin(BaseClass):
    @classmethod
    def get_registry_from_path(
        cls, path: FileStringOrPath, io: IoManager
    ) -> YamlFile | None:
        from wexample_filestate.item.file.yaml_file import YamlFile

        registry_path = cls.get_registry_path_from_path(path=path)
        if registry_path.exists():
            return YamlFile.create_from_path(path=registry_path, io=io)
        return None

    @classmethod
    def get_registry_path_from_path(cls, path: FileStringOrPath) -> FileStringOrPath:
        from pathlib import Path

        from wexample_app.const.globals import WORKDIR_SETUP_DIR
        from wexample_wex_core.const.globals import (
            CORE_DIR_NAME_TMP,
            CORE_FILE_NAME_REGISTRY,
        )

        return (
            Path(path) / WORKDIR_SETUP_DIR / CORE_DIR_NAME_TMP / CORE_FILE_NAME_REGISTRY
        )

    def build_registry_value(self) -> NestedConfigValue:
        from wexample_config.config_value.nested_config_value import NestedConfigValue

        return NestedConfigValue(
            raw={
                "config": self.get_config(),
                "env": self.get_app_env(),
            }
        )

    def get_registry(self, rebuild: bool = False) -> NestedConfigValue:
        registry = self.get_registry_file(rebuild=rebuild)
        return registry.read_config()

    def get_registry_file(self, rebuild: bool = False) -> YamlFile:
        from wexample_filestate.item.file.yaml_file import YamlFile

        registry_path = self.get_registry_path_from_path(path=self.get_path())
        registry = YamlFile.create_from_path(path=registry_path, io=self.io)

        if rebuild or not registry.get_path().exists():
            registry.write_config(self.build_registry_value())

        return registry



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_wex_core.const.globals import COMMAND_TYPE_ADDON
from wexample_wex_core.decorator.command import command
from wexample_wex_core.decorator.middleware import middleware

from wexample_wex_addon_app.middleware.app_middleware import AppMiddleware
from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir
from wexample_wex_addon_app.workdir.framework_packages_suite_workdir import (
    FrameworkPackageSuiteWorkdir,
)

if TYPE_CHECKING:
    from wexample_wex_core.context.execution_context import ExecutionContext


@middleware(middleware=AppMiddleware)
@command(
    type=COMMAND_TYPE_ADDON,
    description="Validate internal dependencies and propagate versions across all packages in the suite.",
)
def app__version__propagate(
    context: ExecutionContext,
    app_workdir: BasicAppWorkdir,
) -> None:
    if isinstance(app_workdir, FrameworkPackageSuiteWorkdir):
        context.kernel.log(
            f"Propagating version of suite's packages {app_workdir.get_project_name()}"
        )
        # Propagate versions
        app_workdir.propagate_packages_versions()
    else:
        context.kernel.log(
            f"Propagating version of single package {app_workdir.get_project_name()}"
        )
        # Propagate versions
        app_workdir.propagate_version()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from wexample_helpers.classes.abstract_method import abstract_method
from wexample_helpers.const.types import PathOrString
from wexample_wex_core.resolver.addon_command_resolver import AddonCommandResolver

from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir

if TYPE_CHECKING:
    from wexample_config.const.types import DictConfig

    from wexample_wex_addon_app.workdir.code_base_workdir import (
        CodeBaseWorkdir,
    )
    from wexample_wex_addon_app.workdir.mixin.with_suite_tree_workdir_mixin import (
        WithSuiteTreeWorkdirMixin,
    )


class FrameworkPackageSuiteWorkdir(BasicAppWorkdir):
    def build_dependencies_map(self) -> dict[str, list[str]]:
        dependencies = {}
        for package in self.get_packages():
            dependencies[package.get_package_name()] = self.filter_local_packages(
                package.get_dependencies_versions().keys()
            )

        return dependencies

    def build_dependencies_stack(
        self,
        package: CodeBaseWorkdir,
        dependency: CodeBaseWorkdir,
        dependencies_map: dict[str, list[str]],
    ) -> list[CodeBaseWorkdir]:
        """Return the declared dependency chain from `package` to `dependency`.

        Deterministic DFS on the local dependencies map. Returns a list of
        package objects [package, ..., dependency] or an empty list if no path exists.
        """
        start = package.get_package_name()
        target = dependency.get_package_name()

        if start == target:
            return [package]

        # Ensure both nodes exist in the map
        nodes = set(dependencies_map.keys()) | {
            d for deps in dependencies_map.values() for d in deps
        }
        if start not in nodes or target not in nodes:
            return []

        visited: set[str] = set()
        stack: list[tuple[str, list[str]]] = [(start, [start])]

        while stack:
            current, path = stack.pop()
            if current in visited:
                continue
            visited.add(current)

            for neighbor in sorted(dependencies_map.get(current, [])):
                if neighbor in visited:
                    continue
                new_path = path + [neighbor]
                if neighbor == target:
                    chain: list[CodeBaseWorkdir] = []
                    for name in new_path:
                        pkg = self.get_package(name)
                        if pkg is not None:
                            chain.append(pkg)
                    if chain and chain[-1].get_package_name() == target:
                        return chain
                stack.append((neighbor, new_path))

        return []

    def build_ordered_dependencies(self) -> list[str]:
        """Return package names ordered leaves -> trunk."""
        return self.topological_order(self.build_dependencies_map())

    # Publication planning helpers
    def compute_packages_to_publish(self) -> list[CodeBaseWorkdir]:
        """Return packages that changed since their last publication tag.

        If a package has no previous tag, it is considered to be published.
        """
        to_publish: list[CodeBaseWorkdir] = []
        for pkg in self.get_packages():
            if pkg.has_changes_since_last_publication_tag():
                to_publish.append(pkg)
        return to_publish

    def filter_local_packages(self, packages: list[str]) -> list[str]:
        """
        Keep only dependencies that are local to this workspace.

        A local dependency is one whose package name matches one of the packages
        discovered by get_packages().
        """
        # Use the dedicated helper to retrieve local package names
        local_names = set(self.get_local_packages_names())
        if not packages:
            return []
        # Return only those present locally, preserve order and remove duplicates
        seen: set[str] = set()
        filtered: list[str] = []
        for name in packages:
            if name in local_names and name not in seen:
                seen.add(name)
                filtered.append(name)
        return filtered

    def get_dependents(self, package: CodeBaseWorkdir) -> list[CodeBaseWorkdir]:
        dependents = []
        for neighbor_package in self.get_packages():
            if neighbor_package.depends_from(package):
                dependents.append(neighbor_package)
        return dependents

    def get_local_packages_names(self) -> list[str]:
        return [p.get_package_name() for p in self.get_packages()]

    def get_ordered_packages(self) -> list[CodeBaseWorkdir]:
        order = self.build_ordered_dependencies()
        by_name = {p.get_package_name(): p for p in self.get_packages()}
        return [by_name[name] for name in order if name in by_name]

    def get_package(self, package_name: str) -> CodeBaseWorkdir | None:
        for package in self.get_packages():
            if package.get_package_name() == package_name:
                return package
        return None

    def get_packages(self) -> list[CodeBaseWorkdir]:
        return self.find_all_by_type(
            class_type=self._get_children_package_workdir_class(), recursive=True
        )

    def get_packages_paths(self) -> list[Path]:
        """Return all resolved package paths that are directories only."""
        config = self.get_config().search("package_suite.location").get_list()

        resolved: list[Path] = []

        for location in (loc.get_str() for loc in config):
            for path in self.get_path().glob(location):
                if path.is_dir():  # ✅ only keep directories
                    resolved.append(path)

        return resolved

    def packages_execute_function(
        self,
        command: callable,
        arguments: list[str] | None = None,
        force: bool = False,
    ) -> None:
        """
        Execute a Python command function (addon command) on all packages.

        This automatically builds the corresponding manager command
        using AddonCommandResolver.

        Example:
            self.packages_execute_function(
                command=app__setup__install,
                arguments=["--env", env],
            )
        """
        resolved_command = AddonCommandResolver.build_command_from_function(
            command_wrapper=command
        )

        self.packages_execute_manager(
            command=resolved_command,
            arguments=arguments,
            force=force,
        )

    def packages_execute_manager(
        self,
        command: str,
        arguments: list[str] | None = None,
        force: bool = False,
    ) -> None:
        """Execute a manager command on all packages."""
        self._packages_execute(
            cmd=[command] + (arguments or []),
            executor_method=BasicAppWorkdir.manager_run_from_path,
            message="Executing command",
            force=force,
        )

    def packages_execute_shell(self, cmd: list[str], force: bool = False) -> None:
        """Execute a raw shell command on all packages."""
        self._packages_execute(
            cmd=cmd,
            executor_method=BasicAppWorkdir.shell_run_from_path,
            message="Executing shell",
            force=force,
        )

    def packages_validate_internal_dependencies_declarations(self) -> None:
        """Ensure imports match declared local dependencies."""
        from wexample_wex_addon_app.exception.dependency_violation_exception import (
            DependencyViolationException,
        )

        dependencies_map = self.build_dependencies_map()

        self.io.log("Checking packages dependencies consistency...")
        self.io.indentation_up()
        progress = self.io.progress(
            total=len(dependencies_map), print_response=False
        ).get_handle()

        for package_name in dependencies_map:
            package = self.get_package(package_name)
            if package is None:
                continue

            search_fn = getattr(package, "search_imports_in_codebase", None)
            if not callable(search_fn):
                progress.advance(
                    label=f"Package {package.get_project_name()} (no search)", step=1
                )
                continue

            for package_name_search in dependencies_map:
                searched_package = self.get_package(package_name_search)
                if searched_package is None:
                    continue

                imports = search_fn(searched_package)
                if len(imports) == 0:
                    continue

                dependencies_stack = self.build_dependencies_stack(
                    package, searched_package, dependencies_map
                )

                if len(dependencies_stack) == 0:
                    import_locations = [
                        f"{res.item.get_path()}:{res.line}:{res.column}"
                        for res in imports
                    ]
                    raise DependencyViolationException(
                        package_name=package_name,
                        imported_package=package_name_search,
                        import_locations=import_locations,
                    )

            progress.advance(label=f"Package {package.get_project_name()}", step=1)

        self.io.success("Internal dependencies match.")
        self.io.indentation_down()

    def prepare_value(self, raw_value: DictConfig | None = None) -> DictConfig:
        """Prepare file state configuration for package suite."""
        raw_value = super().prepare_value(raw_value=raw_value)

        children = raw_value["children"]
        base_path = Path(self.get_path()).resolve()
        package_paths = self.get_packages_paths()

        from wexample_filestate.const.disk import DiskItemType

        # Root of the generated tree
        tree: dict = {
            "name": base_path.name,
            "type": DiskItemType.DIRECTORY,
            "children": {},
        }

        for package_path in package_paths:
            package_path = package_path.resolve()

            # Skip invalid paths
            if not BasicAppWorkdir.is_app_workdir_path(path=package_path):
                continue

            # Build relative path between the configured root and the package path
            rel_parts = list(package_path.relative_to(base_path).parts)

            # Walk the tree structure and create missing nodes
            current = tree
            for part in rel_parts[:-1]:  # all intermediate directories
                current = current["children"].setdefault(
                    part,
                    {
                        "name": part,
                        "type": DiskItemType.DIRECTORY,
                        "children": {},
                    },
                )

            # Add final package directory
            leaf_name = rel_parts[-1]
            current["children"][leaf_name] = {
                "name": leaf_name,
                "class": self._get_children_package_workdir_class(),
                "type": DiskItemType.DIRECTORY,
                "active": False,
                "children": {},  # packages may also contain children later
            }

        # Convert "children" dicts → lists (required format)
        def normalize(node: dict) -> dict:
            if isinstance(node.get("children"), dict):
                node["children"] = [
                    normalize(child) for child in node["children"].values()
                ]
            return node

        # Only append children of the generated root (not the root itself)
        children.extend(normalize(tree)["children"])

        return raw_value

    def propagate_packages_versions(self) -> None:
        for package in self.get_ordered_packages():
            self.propagate_version_of(package=package)

    def propagate_version_of(self, package: WithSuiteTreeWorkdirMixin) -> None:
        package.log(f"Propagating version {package.get_project_version()}", prefix=True)
        package.io.indentation_up()

        for dependent in self.get_dependents(package):
            updated = dependent.save_dependency_from_package(package)
            if updated:
                package.log(
                    f"Updated {dependent.get_project_name()} dependencies",
                )

        package.success("Packages version has been propagated across suite")
        package.io.indentation_down()

    def publish_dependencies(self) -> dict[str, str]:
        """The suite provides dependency of package it manages."""
        dependencies = {}

        for package in self.get_packages():
            dependencies[package.get_package_name()] = package.get_project_version()

        return dependencies

    def setup_install(self, env: str | None = None, force: bool = False) -> None:
        from wexample_wex_addon_app.commands.setup.install import app__setup__install

        self.subtitle(f"Installing suite app")
        super().setup_install(env=env)

        self.subtitle(f"Installing local packages")
        for package in self.get_packages():
            package.ensure_app_manager_setup()

        self.packages_execute_function(
            command=app__setup__install,
            arguments=["--env", env],
        )

    def topological_order(self, dep_map: dict[str, list[str]]) -> list[str]:
        """Deterministic topological order (leaves -> trunk) using graphlib."""
        from graphlib import CycleError, TopologicalSorter

        # Normalize: include every mentioned node and sort for stable results
        nodes = set(dep_map.keys()) | {d for deps in dep_map.values() for d in deps}
        normalized: dict[str, list[str]] = {
            key: sorted([dep for dep in dep_map.get(key, []) if dep in nodes])
            for key in sorted(nodes)
        }

        ts = TopologicalSorter()
        for key, deps in normalized.items():
            ts.add(key, *deps)

        try:
            order = list(ts.static_order())
        except CycleError as err:
            msg = getattr(err, "args", [None])[0] or "Cyclic dependencies detected"
            raise ValueError(str(msg)) from err

        # Return only local packages (original keys of dep_map)
        return [name for name in order if name in dep_map]

    @abstract_method
    def _child_is_package_directory(self, entry: Path) -> bool:
        pass

    def _create_package_workdir(self, package_path: Path) -> CodeBaseWorkdir | None:
        """Create a workdir instance for a package at the given path.

        This method should be overridden by subclasses to return the appropriate
        workdir type (e.g., PythonPackageWorkdir).
        """
        from wexample_filestate.utils.file_state_manager import FileStateManager

        # Get the workdir class from the child implementation
        self._get_children_package_workdir_class()

        # Check if this path is a valid package directory
        if not self._child_is_package_directory(package_path):
            return None

        # Create the workdir instance
        return FileStateManager.create_from_path(
            path=package_path,
            config={},
            io=self.io,
        )

    def _get_children_package_directory_name(self) -> str:
        return "packages"

    def _get_children_package_workdir_class(self) -> type[CodeBaseWorkdir]:
        from wexample_wex_addon_app.workdir.code_base_workdir import (
            CodeBaseWorkdir,
        )

        return CodeBaseWorkdir

    def _package_title(self, path: PathOrString, message: str) -> None:
        from wexample_helpers.helpers.cli import cli_make_clickable_path

        self.title(f"📦 {message}: {path.name}")
        self.log(f"Path: {cli_make_clickable_path(path)}", indentation=1)

    def _packages_execute(
        self,
        cmd: list[str],
        executor_method: callable,
        message: str,
        force: bool = False,
    ) -> None:
        """
        Generic method to execute a command on all detected packages.

        Args:
            cmd: Command to execute (as a list of strings)
            executor_method: Method used to execute the command (e.g. manager_run_from_path, shell_run_from_path)
            message: Displayed title message
            force: If True, run even if directory is not recognized as an app workdir
        """
        import shlex

        from wexample_prompt.enums.terminal_color import TerminalColor

        for package_path in self.get_packages_paths():
            if not force and not BasicAppWorkdir.is_app_workdir_path(path=package_path):
                continue

            self._package_title(path=package_path, message=message)
            self.log(f"Command: {shlex.join(cmd)}", indentation=1)
            self.separator(color=TerminalColor.BLACK)

            try:
                result = executor_method(cmd=cmd, path=package_path)
                if result is None:
                    self.log("Invalid package directory, skipping.", indentation=1)
            except Exception as e:
                self.log(f"Error executing command: {e}", indentation=1)
                return

            self.separator(color=TerminalColor.BLACK)



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_app.response.abstract_response import AbstractResponse
from wexample_helpers.classes.field import public_field
from wexample_helpers.decorator.base_class import base_class
from wexample_prompt.enums.terminal_color import TerminalColor
from wexample_prompt.responses.abstract_prompt_response import AbstractPromptResponse

from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir

if TYPE_CHECKING:
    from wexample_prompt.responses.echo_prompt_response import EchoPromptResponse


@base_class
class AppInfoResponse(AbstractResponse):
    app_workdir: BasicAppWorkdir = public_field(
        description="The application workdir to display information about"
    )

    @staticmethod
    def _format_yes_no(condition: bool) -> str:
        """Format a boolean condition as colored Yes/No.

        Args:
            condition: True for green Yes, False for red No

        Returns:
            Formatted string with color markup
        """
        return "@color:green{Yes}" if condition else "@color:red{No}"

    def _get_coverage_data(self) -> tuple[int, int, int, TerminalColor]:
        """Extract and calculate coverage data from config.

        Returns:
            Tuple of (total, covered, percent, color)
        """
        coverage_last_report = (
            self.app_workdir.get_config()
            .search("test.coverage.last_report")
            .get_dict_or_default()
        )

        if coverage_last_report:
            total = coverage_last_report.get("total")
            covered = coverage_last_report.get("covered")

            total_int = total.get_int() if total else 0
            covered_int = covered.get_int() if covered else 0

            coverage_ratio = (covered_int / total_int) if total_int else 0.0
            clamped_ratio = min(max(coverage_ratio, 0.0), 1.0)
            coverage_percent = int(round(clamped_ratio * 100))
        else:
            total_int = 100
            covered_int = 0
            coverage_percent = 0

        # Determine color based on coverage percentage
        if coverage_percent >= 80:
            coverage_color = TerminalColor.GREEN
        elif coverage_percent > 0:
            coverage_color = TerminalColor.YELLOW
        else:
            coverage_color = TerminalColor.RED

        return total_int, covered_int, coverage_percent, coverage_color

    def _get_formatted_prompt_response(self) -> AbstractPromptResponse:
        """Build the complete app info response with all sections.

        Returns:
            MultiplePromptResponse containing all info sections
        """
        from wexample_app.const.env import ENV_COLORS
        from wexample_prompt.responses.data.multiple_prompt_response import (
            MultiplePromptResponse,
        )
        from wexample_prompt.responses.data.properties_prompt_response import (
            PropertiesPromptResponse,
        )
        from wexample_prompt.responses.interactive.progress_prompt_response import (
            ProgressPromptResponse,
        )
        from wexample_prompt.responses.titles.separator_prompt_response import (
            SeparatorPromptResponse,
        )

        env = self.app_workdir.get_app_env()
        libraries = self._get_libraries_responses()
        total_int, covered_int, coverage_percent, coverage_color = (
            self._get_coverage_data()
        )

        # Check all conditions for publishability
        has_readme = self.app_workdir.has_readme()
        is_clean_since_version = (
            not self.app_workdir.has_changes_since_last_publication_tag()
        )
        has_tests = self.app_workdir.has_a_test()
        is_clean_since_coverage = not self.app_workdir.has_changes_since_last_coverage()

        # App is publishable only if ALL conditions are met
        is_publishable = all(
            [
                has_readme,
                is_clean_since_version,
                has_tests,
                is_clean_since_coverage,
            ]
        )

        # Build response sections
        responses = [
            # Project information
            PropertiesPromptResponse(
                title="Project Info",
                properties={
                    "Name": f"@color:blue{{{self.app_workdir.get_item_name()}}}",
                    "Version": self.app_workdir.get_project_version(),
                    "Path": f"@path{{{self.app_workdir.get_path()}}}",
                    "Environment": f"@color:{ENV_COLORS[env]}{{{env}}}",
                },
            ),
            # Test coverage progress bar
            ProgressPromptResponse(
                total=total_int,
                current=covered_int,
                label=f"Test coverage ({covered_int}/{total_int})",
                color=coverage_color,
                show_percentage=True,
            ),
        ]

        # Add libraries section if any
        if libraries:
            responses.append(
                SeparatorPromptResponse.create_separator(label="Libraries")
            )
            responses.extend(libraries)

        # Add status sections
        responses.extend(
            [
                PropertiesPromptResponse(
                    title="Files",
                    properties={
                        # "Source files": f"@color:magenta{{{self.app_workdir.count_source_files()}}}",
                        # "Test files": f"@color:magenta{{{self.app_workdir.count_test_files()}}}",
                        "Has README.md": self._format_yes_no(has_readme),
                    },
                ),
                PropertiesPromptResponse(
                    title="Code",
                    properties={
                        "Lines in source": f"@color:magenta{{{self.app_workdir.count_source_code_lines()}}}",
                        "Lines in tests": f"@color:magenta{{{self.app_workdir.count_test_code_lines()}}}",
                    },
                ),
                PropertiesPromptResponse(
                    title="Repository",
                    properties={
                        "Clean since last version": self._format_yes_no(
                            is_clean_since_version
                        ),
                    },
                ),
                PropertiesPromptResponse(
                    title="Testing",
                    properties={
                        "Has tests": self._format_yes_no(has_tests),
                        "Clean since last coverage": self._format_yes_no(
                            is_clean_since_coverage
                        ),
                    },
                ),
                PropertiesPromptResponse(
                    title="Status",
                    properties={
                        "Ready to publish": self._format_yes_no(is_publishable),
                    },
                ),
                SeparatorPromptResponse(character="▄"),
            ]
        )

        return MultiplePromptResponse.create_multiple(responses=responses)

    def _get_libraries_responses(self) -> list[EchoPromptResponse]:
        """Get list of library path responses.

        Returns:
            List of EchoPromptResponse for each local library
        """
        from wexample_prompt.responses.echo_prompt_response import EchoPromptResponse

        libraries = []
        local_libraries = self.app_workdir.get_local_libraries_paths()

        if local_libraries:
            for library_config in local_libraries:
                if library_config.is_str():
                    libraries.append(
                        EchoPromptResponse.create_echo(
                            message=f"@path{{{library_config.get_str()}}}"
                        )
                    )

        return libraries



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from wexample_app.const.globals import APP_PATH_APP_MANAGER
from wexample_helpers.decorator.base_class import base_class
from wexample_wex_core.common.abstract_addon_manager import AbstractAddonManager

if TYPE_CHECKING:
    from pathlib import Path

    from wexample_helpers.const.types import PathOrString
    from wexample_wex_core.middleware.abstract_middleware import AbstractMiddleware

    from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir


@base_class
class AppAddonManager(AbstractAddonManager):
    @classmethod
    def get_package_module(cls) -> Any:
        import wexample_wex_addon_app

        return wexample_wex_addon_app

    @classmethod
    def get_shell_manager_path(cls) -> Path:
        from wexample_app.const.globals import APP_FILE_APP_MANAGER

        return (
            cls.get_package_source_path() / "resources" / f"{APP_FILE_APP_MANAGER}.sh"
        )

    def create_app_workdir(
        self, path: PathOrString | None = None
    ) -> BasicAppWorkdir | None:
        from pathlib import Path

        from wexample_helpers.helpers.cli import cli_make_clickable_path
        from wexample_helpers.helpers.module import module_load_class_from_file

        from wexample_wex_addon_app.workdir.basic_app_workdir import BasicAppWorkdir

        app_path = (
            Path(path) if path is not None else self.kernel.call_workdir.get_path()
        )

        if not BasicAppWorkdir.is_app_workdir_path(path=app_path):
            self.kernel.warning(
                f"Path does not match with an application directory structure: {cli_make_clickable_path(app_path)}"
            )
            return None

        custom_app_workdir_class_path = (
            app_path / APP_PATH_APP_MANAGER / "app_workdir.py"
        )
        if custom_app_workdir_class_path.exists():
            app_workdir_class = module_load_class_from_file(
                file_path=custom_app_workdir_class_path, class_name="AppWorkdir"
            )
        else:
            app_workdir_class = BasicAppWorkdir

        # Use basic project class to access minimal configuration.
        return app_workdir_class.create_from_path(
            path=app_path.resolve(),
            parent_io_handler=self.kernel,
        )

    def get_middlewares_classes(self) -> list[type[AbstractMiddleware]]:
        from wexample_wex_addon_app.middleware.app_middleware import AppMiddleware
        from wexample_wex_addon_app.middleware.each_suite_package_middleware import (
            EachSuitePackageMiddleware,
        )
        from wexample_wex_addon_app.middleware.package_suite_middleware import (
            PackageSuiteMiddleware,
        )
        from wexample_wex_addon_app.middleware.suite_or_each_package_middleware import (
            SuiteOrEachPackageMiddleware,
        )

        return [
            AppMiddleware,
            EachSuitePackageMiddleware,
            PackageSuiteMiddleware,
            SuiteOrEachPackageMiddleware,
        ]



from __future__ import annotations

from wexample_app.exception.app_runtime_exception import AppRuntimeException


class InvalidWorkdirTypeException(AppRuntimeException):
    """Exception raised when a workdir is not of the expected type."""

    error_code: str = "INVALID_WORKDIR_TYPE"

    def __init__(
        self,
        workdir_path: str,
        actual_type: str,
        expected_type: str,
        suite_path: str | None = None,
        cause: Exception | None = None,
        previous: Exception | None = None,
    ) -> None:
        from wexample_helpers.helpers.cli import cli_make_clickable_path

        clickable_path = cli_make_clickable_path(workdir_path)

        data = {
            "workdir_path": workdir_path,
            "actual_type": actual_type,
            "expected_type": expected_type,
        }

        message = (
            f"The app_workdir {clickable_path} is of type {actual_type} "
            f"and not a subclass of {expected_type}."
        )

        if suite_path:
            clickable_suite_path = cli_make_clickable_path(suite_path)
            message += f"\nSuite workdir found at: {clickable_suite_path}"
            data["suite_path"] = suite_path

        super().__init__(
            message=message,
            data=data,
            cause=cause,
            previous=previous,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_filestate.config_value.readme_content_config_value import (
    ReadmeContentConfigValue,
)
from wexample_helpers.classes.field import public_field
from wexample_helpers.decorator.base_class import base_class

if TYPE_CHECKING:
    from pathlib import Path


@base_class
class AppReadmeConfigValue(ReadmeContentConfigValue):
    """README generation for applications with workdir and suite support.

    This class handles all common operations for app-based READMEs:
    - Workdir-based template search paths
    - Suite-level template inheritance
    - Predefined section list
    - Common project metadata extraction

    Subclasses only need to implement language-specific methods.
    """

    workdir = public_field(description="The application workdir")

    def _append_template_path_from_module(
        self, module, search_paths: list[str]
    ) -> str | None:
        from wexample_helpers.helpers.module import module_get_path

        """Consider the template directory and the module files are placed at the same relative location into the module directory"""
        template_path = (
            module_get_path(module).parent / "resources" / "readme_templates"
        )

        if template_path.exists():
            search_paths.append(template_path)

    def _get_app_description(self) -> str | None:
        return None

    def _get_app_homepage(self) -> str | None:
        return None

    def _get_dependencies(self) -> dict[str, str]:
        """Extract dependencies from pyproject.toml."""
        return self.workdir.get_dependencies_versions()

    def _get_project_license(self) -> str | None:
        return None

    def _get_readme_search_paths(self) -> list[Path]:
        """Return list of paths to search for README templates.

        Searches in order:
        1. Workdir-specific templates
        2. Suite-level templates (if available)
        3. Bundled default templates

        Returns:
            List of paths to search for templates
        """
        from wexample_app.const.globals import WORKDIR_SETUP_DIR

        workdir_path = self.workdir.get_path()
        search_paths = [
            workdir_path / WORKDIR_SETUP_DIR / "knowledge" / "readme",
        ]

        # Language package specific template
        if __name__ != self.__module__:
            self._append_template_path_from_module(
                module=self.__module__, search_paths=search_paths
            )

        # App package template
        self._append_template_path_from_module(
            module=__name__, search_paths=search_paths
        )

        def _get_template(workdir) -> None:
            search_paths.append(
                workdir.get_path() / WORKDIR_SETUP_DIR / "knowledge" / "package-readme"
            )

        search_paths.extend(
            self.workdir.collect_stack_in_suites_tree(
                callback=_get_template, include_self=False
            )
        )

        return search_paths

    def _get_section_names(self) -> list[str]:
        """Return the list of section names to include in the README.

        Default sections common to all packages. Subclasses can override
        to add language-specific sections.

        Returns:
            List of section names in order
        """
        return super()._get_section_names() + [
            "status-compatibility",
            "prerequisites",
            "installation",
            "quickstart",
            "basic-usage",
            "configuration",
            "logging",
            "api-reference",
            "examples",
            "tests",
            "code-quality",
            "versioning",
            "changelog",
            "migration-notes",
            "roadmap",
            "troubleshooting",
            "security",
            "privacy",
            "support",
            "contribution-guidelines",
            "maintainers",
            "license",
            "useful-links",
            "suite-integration",
            "compatibility-matrix",
            "requirements",
            "dependencies",
            "links",
            "suite-signature",
        ]

    def _get_template_context(self) -> dict:
        """Build the template context with all variables.

        Common variables for all app-based READMEs. Subclasses should
        override and call super() to add language-specific variables.

        Returns:
            Dictionary of template variables
        """
        return {
            # filestate: python-iterable-sort
            "dependencies": self._get_dependencies(),
            "description": self._get_app_description(),
            "homepage": self._get_app_homepage(),
            "license_info": self._get_project_license(),
            "package_name": self.workdir.get_package_name(),
            "project_name": self.workdir.get_project_name(),
            "version": self.workdir.get_project_version(),
            "workdir": self.workdir,
        }



from __future__ import annotations

from wexample_filestate.config_value.aggregated_templates_config_value import (
    AggregatedTemplatesConfigValue,
)
from wexample_helpers.classes.field import public_field
from wexample_helpers.decorator.base_class import base_class
from wexample_wex_addon_app.workdir.code_base_workdir import CodeBaseWorkdir


@base_class
class AppLicenseConfigValue(AggregatedTemplatesConfigValue):
    workdir: CodeBaseWorkdir = public_field(description="The application workdir")

    def get_templates(self) -> list[str] | None:
        from datetime import datetime

        return [
            "MIT License",
            f"Copyright (c) {datetime.now().year} {self.workdir.get_vendor_name()}",
            "",
            "Permission is hereby granted, free of charge, to any person obtaining a copy",
            'of this software and associated documentation files (the "Software"), to deal',
            "in the Software without restriction, including without limitation the rights",
            "to use, copy, modify, merge, publish, distribute, sublicense, and/or sell",
            "copies of the Software, and to permit persons to whom the Software is",
            "furnished to do so, subject to the following conditions:",
            "",
            "The above copyright notice and this permission notice shall be included in",
            "all copies or substantial portions of the Software.",
            "",
            'THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR',
            "IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,",
            "FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE",
            "AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER",
            "LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,",
            "OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE",
            "SOFTWARE.",
            "",
        ]



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_app.const.output import OUTPUT_FORMAT_JSON, OUTPUT_TARGET_FILE
from wexample_app.helpers.request import request_build_id
from wexample_app.item.file.iml_file import ImlFile
from wexample_app.workdir.mixin.with_runtime_config_mixin import WithRuntimeConfigMixin
from wexample_helpers.const.types import FileStringOrPath
from wexample_helpers.decorator.base_class import base_class
from wexample_wex_core.common.app_manager_shell_result import AppManagerShellResult
from wexample_wex_core.resolver.addon_command_resolver import AddonCommandResolver
from wexample_wex_core.workdir.mixin.with_app_version_workdir_mixin import (
    WithAppVersionWorkdirMixin,
)

from wexample_wex_addon_app.workdir.mixin.with_app_config_workdir_mixin import (
    WithAppConfigWorkdirMixin,
)
from wexample_wex_addon_app.workdir.mixin.with_app_registry_mixin import (
    WithAppRegistryMixin,
)
from wexample_wex_addon_app.workdir.mixin.with_readme_workdir_mixin import (
    WithReadmeWorkdirMixin,
)
from wexample_wex_addon_app.workdir.mixin.with_suite_tree_workdir_mixin import (
    WithSuiteTreeWorkdirMixin,
)

if TYPE_CHECKING:
    from wexample_config.const.types import DictConfig
    from wexample_helpers.classes.shell_result import ShellResult


@base_class
class AppWorkdirMixin(
    WithAppConfigWorkdirMixin,
    WithSuiteTreeWorkdirMixin,
    WithReadmeWorkdirMixin,
    WithAppVersionWorkdirMixin,
    WithRuntimeConfigMixin,
    WithAppRegistryMixin,
):
    @classmethod
    def is_app_workdir_path(cls, path: FileStringOrPath) -> bool:
        config = cls.get_config_from_path(path=path)
        if config:
            return not config.read_config().search("global.version").is_none()
        return False

    @classmethod
    def is_app_workdir_path_setup(cls, path: FileStringOrPath) -> bool:
        from pathlib import Path

        from wexample_app.const.globals import APP_PATH_BIN_APP_MANAGER

        path = Path(path)
        if cls.is_app_workdir_path(path=path):
            # app-manager exists.
            return (path / APP_PATH_BIN_APP_MANAGER).exists()
        return False

    @classmethod
    def manager_run_command_from_path(
        cls,
        path: str,
        command: callable,
        arguments: list[str] | None = None,
    ) -> AppManagerShellResult:
        """
        Execute a Python addon command (e.g., app__setup__install) using the app manager,
        within a specific workdir.
        """
        from wexample_app.const.globals import APP_PATH_BIN_APP_MANAGER
        from wexample_helpers.helpers.shell import shell_run

        # Resolve function to CLI command name
        resolved_command = AddonCommandResolver.build_command_from_function(
            command_wrapper=command
        )

        request_id = request_build_id()

        # Build full command
        cmd = [resolved_command] + (arguments or [])
        full_cmd = [
            str(APP_PATH_BIN_APP_MANAGER),
            "--force-request-id",
            request_id,
            "--output-format",
            OUTPUT_FORMAT_JSON,
            "--output-target",
            OUTPUT_TARGET_FILE,
        ] + cmd

        # Run the manager command in the given workdir
        return AppManagerShellResult.from_shell_result(
            request_id=request_id,
            result=shell_run(cmd=full_cmd, cwd=path, inherit_stdio=True),
        )

    @classmethod
    def manager_run_from_path(
        cls, path: FileStringOrPath, cmd: list[str] | str
    ) -> None | ShellResult:
        from wexample_app.const.globals import APP_PATH_BIN_APP_MANAGER
        from wexample_helpers.helpers.shell import shell_run

        if not isinstance(cmd, list):
            cmd = [cmd]

        full_cmd = [str(APP_PATH_BIN_APP_MANAGER)]
        full_cmd.extend(cmd)

        # Ask parent suite to generate the info registry that contains packages readme information
        return shell_run(
            cmd=full_cmd,
            cwd=path,
            inherit_stdio=True,
        )

    @classmethod
    def shell_run_from_path(
        cls, path: FileStringOrPath, cmd: list[str] | str
    ) -> None | ShellResult:
        from wexample_helpers.helpers.shell import shell_run

        return shell_run(
            cmd=cmd,
            cwd=str(path),
            inherit_stdio=True,
        )

    def get_project_name(self) -> str:
        from wexample_app.const.globals import APP_FILE_APP_CONFIG

        name_config = self.get_runtime_config().search("global.name")
        # Ensure we properly handle missing or empty name
        name: str | None = None
        if not name_config.is_none():
            name = (name_config.get_str_or_none() or "").strip()
        # Enforce that a project must have a non-empty name; include path for debug
        if not name:
            raise ValueError(
                f"Project at '{self.get_path()}' must define a non-empty 'global.name' in {APP_FILE_APP_CONFIG}."
            )
        return name

    def get_project_version(self) -> str:
        from wexample_app.const.globals import APP_FILE_APP_CONFIG

        # Ensure we properly handle missing node and empty value
        config = self.get_runtime_config()

        version_config = config.search("global.version")
        version = version_config.get_str_or_none()
        if version is None or str(version).strip() == "":
            raise ValueError(
                f"Project at '{self.get_path()}' must define a non-empty 'version' number in {APP_FILE_APP_CONFIG}."
            )
        return str(version).strip()

    def manager_run_command(self, **kwargs) -> AppManagerShellResult:
        return self.manager_run_command_from_path(path=self.get_path(), **kwargs)

    def prepare_value(self, raw_value: DictConfig | None = None) -> DictConfig:
        from wexample_app.const.globals import (
            APP_FILE_APP_CONFIG,
            APP_FILE_APP_MANAGER,
            WORKDIR_SETUP_DIR,
        )
        from wexample_filestate.config_value.file_content_config_value import (
            FileContentConfigValue,
        )
        from wexample_filestate.const.disk import DiskItemType
        from wexample_filestate.item.file.env_file import EnvFile
        from wexample_filestate.item.file.yaml_file import YamlFile
        from wexample_filestate.option.text_option import TextOption
        from wexample_wex_core.const.globals import CORE_DIR_NAME_TMP
        from wexample_wex_core.const.project import PROJECT_GITIGNORE_DEFAULT

        from wexample_wex_addon_app.app_addon_manager import AppAddonManager

        raw_value = super().prepare_value(raw_value)

        raw_value.update({"mode": {"permissions": "777", "recursive": True}})

        self.append_readme(config=raw_value)
        self.append_version(config=raw_value)

        raw_value["children"].append(
            {
                # .wex
                "name": WORKDIR_SETUP_DIR,
                "type": DiskItemType.DIRECTORY,
                "should_exist": True,
                "children": [
                    {
                        # .env
                        "class": EnvFile,
                        "name": EnvFile.EXTENSION_DOT_ENV,
                        "type": DiskItemType.FILE,
                        "should_exist": True,
                        TextOption.get_name(): {"end_new_line": True},
                    },
                    {
                        # config.yml
                        "name": APP_FILE_APP_CONFIG,
                        "type": DiskItemType.FILE,
                        "should_exist": True,
                        "class": YamlFile,
                        TextOption.get_name(): {"end_new_line": True},
                        "yaml": {"sort_recursive": True},
                    },
                    {
                        # python (app manager)
                        "name": "bin",
                        "type": DiskItemType.DIRECTORY,
                        "should_exist": True,
                        "children": [
                            {
                                "name": APP_FILE_APP_MANAGER,
                                "type": DiskItemType.FILE,
                                "should_exist": True,
                                "content": FileContentConfigValue(
                                    path=AppAddonManager.get_shell_manager_path()
                                ),
                            }
                        ],
                    },
                    {
                        # tmp
                        "name": CORE_DIR_NAME_TMP,
                        "type": DiskItemType.DIRECTORY,
                        "should_exist": True,
                    },
                    {
                        "name": ".gitignore",
                        "type": DiskItemType.FILE,
                        "should_exist": True,
                        "should_contain_lines": [EnvFile.EXTENSION_DOT_ENV],
                        TextOption.get_name(): {"end_new_line": True},
                    },
                ],
            }
        )

        raw_value["children"].append(
            {
                "name": ".gitignore",
                "type": DiskItemType.FILE,
                "should_exist": True,
                TextOption.get_name(): {"end_new_line": True},
                "should_contain_lines": PROJECT_GITIGNORE_DEFAULT,
            }
        )

        raw_value["children"].append(
            {
                # .iml
                "class": self._get_iml_file_class(),
                "name": ImlFile.get_dotted_extension(),
                "type": DiskItemType.FILE,
                "should_exist": True,
            },
        )

        return raw_value

    def shell_run_for_app(self, **kwargs) -> ShellResult:
        return self.shell_run_from_path(path=self.get_path(), **kwargs)

    def _get_iml_file_class(self) -> type[ImlFile]:
        return ImlFile
